##################### generated by xml-casa (v2) from calibeovsa.xml ################
##################### f295775a00a77dd77b81d6e69924145c ##############################
from __future__ import absolute_import
from casashell.private.stack_manip import find_local as __sf__
from casashell.private.stack_manip import find_frame as _find_frame
from casatools.typecheck import validator as _pc
from casatools.coercetype import coerce as _coerce
from suncasatasks import calibeovsa as _calibeovsa_t
from collections import OrderedDict
import numpy
import sys
import os

import shutil

def static_var(varname, value):
    def decorate(func):
        setattr(func, varname, value)
        return func
    return decorate

class _calibeovsa:
    """
    calibeovsa ---- Calibrating EOVSA one or more measurement sets using calibration products in the SQL database.

    
    Calibrating EOVSA one or more measurement sets using calibration products in the SQL database. This task currently only works on pipeline.
    

    --------- parameter descriptions ---------------------------------------------

    vis          input EOVSA (uncalibrated) measurement set(s).
    caltype      Types of calibrations to perform
    caltbdir     Directory to place calibration tables.
    interp       Temporal interpolation for phacal table(s) (nearest or linear)
    docalib      If False, only create the calibration tables but do not perform applycal.
    doflag       If true then perform flagging.
    flagant      Antennas to be flagged. Follow CASA syntax of "antenna".
    doimage      If True, produce a quicklook image after calibration (sunpy must be installed).
    imagedir     directory to place output images. Default current directory.
    antenna      antenna/baselines to be used for imaging. Follow CASA syntax of "antenna".
    timerange    Timerange to be imaged. Follow CASA syntax of "timerange". Default is the entire duration of the ms.
    spw          spectral windows to be imaged. Follow CASA syntax of "spw".
    stokes       stokes to be imaged. Follow CASA syntax of "stokes".
    dosplit      If True, plit the corrected data column as output visibility file.
    outputvis    Name of output visibility file. Default is the name of the first vis file ended with ".corrected.ms".
    doconcat     If True, and if more than one visibility dataset provided, concatenate all into one visibility.
    concatvis    Name of output visibility file. Default is the name of the first + last vis file ended with ".corrected.ms".
    keep_orig_ms Keep the original seperated ms datasets after split?

    --------- examples -----------------------------------------------------------

    
    Calibrating EOVSA one or more measurement sets using calibration products in the SQL database.
    
    Detailed Keyword arguments:
    
    vis -- Name of input EOVSA measurement set dataset(s)
    default: none. Must be supplied
    example: vis = 'IDB20160524000518.ms'
    example: vis = ['IDB20160524000518.ms','IDB20160524000528.ms']
    
    caltype -- list. Type of calibrations to be applied.
    'refpha': reference phase calibration
    'refamp': reference amplitude calibration (not used anymore)
    'phacal': daily phase calibration
    'fluxcal': flux calibration based on total-power measurements
    default value: ['refpha','phacal']
    *** note fluxcal is already implemented in udb_corr when doing importeovsa, should not be used anymore ****
    *** pipeline only uses ['refpha','phacal']
    
    caltbdir -- string. Place to hold calibration tables. Default is current directory. Pipeline should use /data1/eovsa/caltable
    
    interp -- string. How interpolation is done for phacal? 'nearest' or 'linear'
    
    docalib -- boolean. Default True. If False, only create the calibration tables but do not perform applycal
    
    doflag -- boolean. Default True. Peforming flags?
    
    flagant -- string. Follow CASA antenna selection syntax. Default '13~15'.
    
    doimage -- boolean. Default False. If true, make a quicklook image using the specified time range and specified spw range
    
    imagedir -- string. Directory to place the output image.
    
    antenna -- string. Default '0~12'. Antenna/baselines to be used for imaging. Follow CASA antenna selection syntax.
    
    timerange -- string. Default '' (the whole duration of the visibility data). Follow CASA timerange syntax.
    e.g., '2017/07/11/20:16:00~2017/07/11/20:17:00'
    
    spw -- string. Default '1~3'. Follow CASA spw selection syntax.
    
    stokes -- string. Which stokes for the quicklook image. CASA syntax. Default 'XX'
    
    dosplit -- boolean. Split the corrected data column?
    
    outputvis -- string. Output visibility file after split
    
    doconcat -- boolean. If more than one visibility dataset provided, concatenate all into one or make separate outputs if True
    
    concatvis -- string. Output visibility file after concatenation
    
    keep_orig_ms -- boolean. Default True. Inherited from suncasa.eovsa.concateovsa.
    Keep the original seperated ms datasets after concatenation?
    
    


    """

    _info_group_ = """Calibration"""
    _info_desc_ = """Calibrating EOVSA one or more measurement sets using calibration products in the SQL database."""

    __schema = {'vis': {'anyof': [{'type': 'cStr', 'coerce': _coerce.to_str}, {'type': 'cStrVec', 'coerce': [_coerce.to_list,_coerce.to_strvec]}]}, 'caltype': {'anyof': [{'type': 'cStr', 'coerce': _coerce.to_str}, {'type': 'cStrVec', 'coerce': [_coerce.to_list,_coerce.to_strvec]}]}, 'caltbdir': {'type': 'cStr', 'coerce': _coerce.to_str}, 'interp': {'type': 'cStr', 'coerce': _coerce.to_str}, 'docalib': {'type': 'cBool'}, 'doflag': {'type': 'cBool'}, 'flagant': {'type': 'cStr', 'coerce': _coerce.to_str}, 'doimage': {'type': 'cBool'}, 'imagedir': {'type': 'cStr', 'coerce': _coerce.to_str}, 'antenna': {'type': 'cStr', 'coerce': _coerce.to_str}, 'timerange': {'type': 'cStr', 'coerce': _coerce.to_str}, 'spw': {'type': 'cStr', 'coerce': _coerce.to_str}, 'stokes': {'type': 'cStr', 'coerce': _coerce.to_str}, 'dosplit': {'type': 'cBool'}, 'outputvis': {'anyof': [{'type': 'cStr', 'coerce': _coerce.to_str}, {'type': 'cStrVec', 'coerce': [_coerce.to_list,_coerce.to_strvec]}]}, 'doconcat': {'type': 'cBool'}, 'concatvis': {'type': 'cStr', 'coerce': _coerce.to_str}, 'keep_orig_ms': {'type': 'cBool'}}

    def __init__(self):
        self.__stdout = None
        self.__stderr = None
        self.__root_frame_ = None

    def __globals_(self):
        if self.__root_frame_ is None:
            self.__root_frame_ = _find_frame( )
            assert self.__root_frame_ is not None, "could not find CASAshell global frame"
        return self.__root_frame_

    def __to_string_(self,value):
        if type(value) is str:
            return "'%s'" % value
        else:
            return str(value)

    def __validate_(self,doc,schema):
        return _pc.validate(doc,schema)

    def __do_inp_output(self,param_prefix,description_str,formatting_chars):
        out = self.__stdout or sys.stdout
        description = description_str.split( )
        prefix_width = 23 + 15 + 4
        output = [ ]
        addon = ''
        first_addon = True
        if len(description) == 0:
            out.write(param_prefix + " #\n")
            return
        while len(description) > 0:
            ## starting a new line.....................................................................
            if len(output) == 0:
                ## for first line add parameter information............................................
                if len(param_prefix)-formatting_chars > prefix_width - 1:
                    output.append(param_prefix)
                    continue
                addon = param_prefix + ' #'
                first_addon = True
                addon_formatting = formatting_chars
            else:
                ## for subsequent lines space over prefix width........................................
                addon = (' ' * prefix_width) + '#'
                first_addon = False
                addon_formatting = 0
            ## if first word of description puts us over the screen width, bail........................
            if len(addon + description[0]) - addon_formatting + 1 > self.term_width:
                ## if we're doing the first line make sure it's output.................................
                if first_addon: output.append(addon)
                break
            while len(description) > 0:
                ## if the next description word puts us over break for the next line...................
                if len(addon + description[0]) - addon_formatting + 1 > self.term_width: break
                addon = addon + ' ' + description[0]
                description.pop(0)
            output.append(addon)
        out.write('\n'.join(output) + '\n')

    #--------- return nonsubparam values ----------------------------------------------

    def __doimage_dflt( self, glb ):
        return False

    def __doimage( self, glb ):
        if 'doimage' in glb: return glb['doimage']
        return False

    def __vis_dflt( self, glb ):
        return ''

    def __vis( self, glb ):
        if 'vis' in glb: return glb['vis']
        return ''

    def __caltbdir_dflt( self, glb ):
        return ''

    def __caltbdir( self, glb ):
        if 'caltbdir' in glb: return glb['caltbdir']
        return ''

    def __docalib_dflt( self, glb ):
        return True

    def __docalib( self, glb ):
        if 'docalib' in glb: return glb['docalib']
        return True

    def __interp_dflt( self, glb ):
        return 'nearest'

    def __interp( self, glb ):
        if 'interp' in glb: return glb['interp']
        return 'nearest'

    def __caltype_dflt( self, glb ):
        return [  ]

    def __caltype( self, glb ):
        if 'caltype' in glb: return glb['caltype']
        return [  ]

    def __doflag_dflt( self, glb ):
        return True

    def __doflag( self, glb ):
        if 'doflag' in glb: return glb['doflag']
        return True

    def __dosplit_dflt( self, glb ):
        return False

    def __dosplit( self, glb ):
        if 'dosplit' in glb: return glb['dosplit']
        return False

    def __doconcat_dflt( self, glb ):
        return False

    def __doconcat( self, glb ):
        if 'doconcat' in glb: return glb['doconcat']
        return False



    #--------- return inp/go default --------------------------------------------------
    def __antenna_dflt( self, glb ):
        if self.__doimage( glb ) == bool(True): return "0~12"
        return None
    def __stokes_dflt( self, glb ):
        if self.__doimage( glb ) == bool(True): return "XX"
        return None
    def __flagant_dflt( self, glb ):
        if self.__doflag( glb ) == bool(True): return "13~15"
        return None
    def __concatvis_dflt( self, glb ):
        if self.__doconcat( glb ) == bool(True): return ""
        return None
    def __outputvis_dflt( self, glb ):
        if self.__dosplit( glb ) == bool(True): return ""
        return None
    def __keep_orig_ms_dflt( self, glb ):
        if self.__dosplit( glb ) == bool(True): return bool(True)
        if self.__doconcat( glb ) == bool(True): return bool(True)
        return None
    def __imagedir_dflt( self, glb ):
        if self.__doimage( glb ) == bool(True): return "."
        return None
    def __spw_dflt( self, glb ):
        if self.__doimage( glb ) == bool(True): return "1~3"
        return None
    def __timerange_dflt( self, glb ):
        if self.__doimage( glb ) == bool(True): return ""
        return None

    #--------- return subparam values -------------------------------------------------
    def __flagant( self, glb ):
        if 'flagant' in glb: return glb['flagant']
        dflt = self.__flagant_dflt( glb )
        if dflt is not None: return dflt
        return '13~15'
    def __imagedir( self, glb ):
        if 'imagedir' in glb: return glb['imagedir']
        dflt = self.__imagedir_dflt( glb )
        if dflt is not None: return dflt
        return '.'
    def __antenna( self, glb ):
        if 'antenna' in glb: return glb['antenna']
        dflt = self.__antenna_dflt( glb )
        if dflt is not None: return dflt
        return '0~12'
    def __timerange( self, glb ):
        if 'timerange' in glb: return glb['timerange']
        dflt = self.__timerange_dflt( glb )
        if dflt is not None: return dflt
        return ''
    def __spw( self, glb ):
        if 'spw' in glb: return glb['spw']
        dflt = self.__spw_dflt( glb )
        if dflt is not None: return dflt
        return '1~3'
    def __stokes( self, glb ):
        if 'stokes' in glb: return glb['stokes']
        dflt = self.__stokes_dflt( glb )
        if dflt is not None: return dflt
        return 'XX'
    def __outputvis( self, glb ):
        if 'outputvis' in glb: return glb['outputvis']
        dflt = self.__outputvis_dflt( glb )
        if dflt is not None: return dflt
        return ''
    def __concatvis( self, glb ):
        if 'concatvis' in glb: return glb['concatvis']
        dflt = self.__concatvis_dflt( glb )
        if dflt is not None: return dflt
        return ''
    def __keep_orig_ms( self, glb ):
        if 'keep_orig_ms' in glb: return glb['keep_orig_ms']
        dflt = self.__keep_orig_ms_dflt( glb )
        if dflt is not None: return dflt
        return True

    #--------- subparam inp output ----------------------------------------------------
    def __vis_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__vis_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return ''
        description = ''
        value = self.__vis( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'vis': value},{'vis': self.__schema['vis']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-15.15s = %s%-23s%s' % ('vis',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __caltype_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__caltype_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return [  ]
        description = ''
        value = self.__caltype( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'caltype': value},{'caltype': self.__schema['caltype']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-15.15s = %s%-23s%s' % ('caltype',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __caltbdir_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__caltbdir_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return ''
        description = ''
        value = self.__caltbdir( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'caltbdir': value},{'caltbdir': self.__schema['caltbdir']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-15.15s = %s%-23s%s' % ('caltbdir',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __interp_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__interp_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return 'nearest'
        description = ''
        value = self.__interp( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'interp': value},{'interp': self.__schema['interp']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-15.15s = %s%-23s%s' % ('interp',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __docalib_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__docalib_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return True
        description = ''
        value = self.__docalib( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'docalib': value},{'docalib': self.__schema['docalib']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-15.15s = %s%-23s%s' % ('docalib',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __doflag_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__doflag_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return True
        description = ''
        value = self.__doflag( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'doflag': value},{'doflag': self.__schema['doflag']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('\x1B[1m\x1B[47m%-15.15s =\x1B[0m %s%-23s%s' % ('doflag',pre,self.__to_string_(value),post),description,13+len(pre)+len(post))
    def __flagant_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__flagant_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return '13~15'
        if self.__flagant_dflt( self.__globals_( ) ) is not None:
             description = ''
             value = self.__flagant( self.__globals_( ) )
             (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'flagant': value},{'flagant': self.__schema['flagant']}) else ('\x1B[91m','\x1B[0m')
             self.__do_inp_output('   \x1B[92m%-12.12s =\x1B[0m %s%-23s%s' % ('flagant',pre,self.__to_string_(value),post),description,9+len(pre)+len(post))
    def __doimage_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__doimage_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return False
        description = ''
        value = self.__doimage( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'doimage': value},{'doimage': self.__schema['doimage']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('\x1B[1m\x1B[47m%-15.15s =\x1B[0m %s%-23s%s' % ('doimage',pre,self.__to_string_(value),post),description,13+len(pre)+len(post))
    def __imagedir_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__imagedir_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return '.'
        if self.__imagedir_dflt( self.__globals_( ) ) is not None:
             description = ''
             value = self.__imagedir( self.__globals_( ) )
             (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'imagedir': value},{'imagedir': self.__schema['imagedir']}) else ('\x1B[91m','\x1B[0m')
             self.__do_inp_output('   \x1B[92m%-12.12s =\x1B[0m %s%-23s%s' % ('imagedir',pre,self.__to_string_(value),post),description,9+len(pre)+len(post))
    def __antenna_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__antenna_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return '0~12'
        if self.__antenna_dflt( self.__globals_( ) ) is not None:
             description = ''
             value = self.__antenna( self.__globals_( ) )
             (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'antenna': value},{'antenna': self.__schema['antenna']}) else ('\x1B[91m','\x1B[0m')
             self.__do_inp_output('   \x1B[92m%-12.12s =\x1B[0m %s%-23s%s' % ('antenna',pre,self.__to_string_(value),post),description,9+len(pre)+len(post))
    def __timerange_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__timerange_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return ''
        if self.__timerange_dflt( self.__globals_( ) ) is not None:
             description = ''
             value = self.__timerange( self.__globals_( ) )
             (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'timerange': value},{'timerange': self.__schema['timerange']}) else ('\x1B[91m','\x1B[0m')
             self.__do_inp_output('   \x1B[92m%-12.12s =\x1B[0m %s%-23s%s' % ('timerange',pre,self.__to_string_(value),post),description,9+len(pre)+len(post))
    def __spw_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__spw_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return '1~3'
        if self.__spw_dflt( self.__globals_( ) ) is not None:
             description = ''
             value = self.__spw( self.__globals_( ) )
             (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'spw': value},{'spw': self.__schema['spw']}) else ('\x1B[91m','\x1B[0m')
             self.__do_inp_output('   \x1B[92m%-12.12s =\x1B[0m %s%-23s%s' % ('spw',pre,self.__to_string_(value),post),description,9+len(pre)+len(post))
    def __stokes_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__stokes_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return 'XX'
        if self.__stokes_dflt( self.__globals_( ) ) is not None:
             description = ''
             value = self.__stokes( self.__globals_( ) )
             (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'stokes': value},{'stokes': self.__schema['stokes']}) else ('\x1B[91m','\x1B[0m')
             self.__do_inp_output('   \x1B[92m%-12.12s =\x1B[0m %s%-23s%s' % ('stokes',pre,self.__to_string_(value),post),description,9+len(pre)+len(post))
    def __dosplit_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__dosplit_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return False
        description = ''
        value = self.__dosplit( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'dosplit': value},{'dosplit': self.__schema['dosplit']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('\x1B[1m\x1B[47m%-15.15s =\x1B[0m %s%-23s%s' % ('dosplit',pre,self.__to_string_(value),post),description,13+len(pre)+len(post))
    def __outputvis_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__outputvis_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return ''
        if self.__outputvis_dflt( self.__globals_( ) ) is not None:
             description = ''
             value = self.__outputvis( self.__globals_( ) )
             (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'outputvis': value},{'outputvis': self.__schema['outputvis']}) else ('\x1B[91m','\x1B[0m')
             self.__do_inp_output('   \x1B[92m%-12.12s =\x1B[0m %s%-23s%s' % ('outputvis',pre,self.__to_string_(value),post),description,9+len(pre)+len(post))
    def __doconcat_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__doconcat_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return False
        description = ''
        value = self.__doconcat( self.__globals_( ) )
        (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'doconcat': value},{'doconcat': self.__schema['doconcat']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('\x1B[1m\x1B[47m%-15.15s =\x1B[0m %s%-23s%s' % ('doconcat',pre,self.__to_string_(value),post),description,13+len(pre)+len(post))
    def __concatvis_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__concatvis_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return ''
        if self.__concatvis_dflt( self.__globals_( ) ) is not None:
             description = ''
             value = self.__concatvis( self.__globals_( ) )
             (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'concatvis': value},{'concatvis': self.__schema['concatvis']}) else ('\x1B[91m','\x1B[0m')
             self.__do_inp_output('   \x1B[92m%-12.12s =\x1B[0m %s%-23s%s' % ('concatvis',pre,self.__to_string_(value),post),description,9+len(pre)+len(post))
    def __keep_orig_ms_inp(self):
        def xml_default( ):
            ## play the crazy subparameter shell game
            dflt = self.__keep_orig_ms_dflt( self.__globals_( ) )
            if dflt is not None: return dflt
            return True
        if self.__keep_orig_ms_dflt( self.__globals_( ) ) is not None:
             description = ''
             value = self.__keep_orig_ms( self.__globals_( ) )
             (pre,post) = (('','') if value == xml_default( ) else ('\x1B[34m','\x1B[0m')) if self.__validate_({'keep_orig_ms': value},{'keep_orig_ms': self.__schema['keep_orig_ms']}) else ('\x1B[91m','\x1B[0m')
             self.__do_inp_output('   \x1B[92m%-12.12s =\x1B[0m %s%-23s%s' % ('keep_orig_ms',pre,self.__to_string_(value),post),description,9+len(pre)+len(post))

    #--------- global default implementation-------------------------------------------
    @static_var('state', __sf__('casa_inp_go_state'))
    def set_global_defaults(self):
        self.set_global_defaults.state['last'] = self
        glb = self.__globals_( )
        if 'antenna' in glb: del glb['antenna']
        if 'stokes' in glb: del glb['stokes']
        if 'flagant' in glb: del glb['flagant']
        if 'concatvis' in glb: del glb['concatvis']
        if 'dosplit' in glb: del glb['dosplit']
        if 'doflag' in glb: del glb['doflag']
        if 'outputvis' in glb: del glb['outputvis']
        if 'keep_orig_ms' in glb: del glb['keep_orig_ms']
        if 'vis' in glb: del glb['vis']
        if 'interp' in glb: del glb['interp']
        if 'caltype' in glb: del glb['caltype']
        if 'docalib' in glb: del glb['docalib']
        if 'doconcat' in glb: del glb['doconcat']
        if 'imagedir' in glb: del glb['imagedir']
        if 'doimage' in glb: del glb['doimage']
        if 'caltbdir' in glb: del glb['caltbdir']
        if 'spw' in glb: del glb['spw']
        if 'timerange' in glb: del glb['timerange']


    #--------- inp function -----------------------------------------------------------
    def inp(self):
        print("# calibeovsa -- %s" % self._info_desc_)
        self.term_width, self.term_height = shutil.get_terminal_size(fallback=(80, 24))
        self.__vis_inp( )
        self.__caltype_inp( )
        self.__caltbdir_inp( )
        self.__interp_inp( )
        self.__docalib_inp( )
        self.__doflag_inp( )
        self.__flagant_inp( )
        self.__doimage_inp( )
        self.__imagedir_inp( )
        self.__antenna_inp( )
        self.__timerange_inp( )
        self.__spw_inp( )
        self.__stokes_inp( )
        self.__dosplit_inp( )
        self.__outputvis_inp( )
        self.__doconcat_inp( )
        self.__concatvis_inp( )
        self.__keep_orig_ms_inp( )

    #--------- tget function ----------------------------------------------------------
    @static_var('state', __sf__('casa_inp_go_state'))
    def tget(self,savefile=None):
        from casashell.private.stack_manip import find_frame
        from runpy import run_path
        filename = savefile
        if filename is None:
            filename = "calibeovsa.last" if os.path.isfile("calibeovsa.last") else "calibeovsa.saved"
        if os.path.isfile(filename):
            glob = find_frame( )
            newglob = run_path( filename, init_globals={ } )
            for i in newglob:
                glob[i] = newglob[i]
            self.tget.state['last'] = self
        else:
            print("could not find last file: %s\nsetting defaults instead..." % filename)
            self.set_global_defaults( )

    #--------- tput function ----------------------------------------------------------
    def tput(self,outfile=None):
        def noobj(s):
           if s.startswith('<') and s.endswith('>'):
               return "None"
           else:
               return s

        _postfile = outfile if outfile is not None else os.path.realpath('calibeovsa.last')

        _invocation_parameters = OrderedDict( )
        _invocation_parameters['vis'] = self.__vis( self.__globals_( ) )
        _invocation_parameters['caltype'] = self.__caltype( self.__globals_( ) )
        _invocation_parameters['caltbdir'] = self.__caltbdir( self.__globals_( ) )
        _invocation_parameters['interp'] = self.__interp( self.__globals_( ) )
        _invocation_parameters['docalib'] = self.__docalib( self.__globals_( ) )
        _invocation_parameters['doflag'] = self.__doflag( self.__globals_( ) )
        _invocation_parameters['flagant'] = self.__flagant( self.__globals_( ) )
        _invocation_parameters['doimage'] = self.__doimage( self.__globals_( ) )
        _invocation_parameters['imagedir'] = self.__imagedir( self.__globals_( ) )
        _invocation_parameters['antenna'] = self.__antenna( self.__globals_( ) )
        _invocation_parameters['timerange'] = self.__timerange( self.__globals_( ) )
        _invocation_parameters['spw'] = self.__spw( self.__globals_( ) )
        _invocation_parameters['stokes'] = self.__stokes( self.__globals_( ) )
        _invocation_parameters['dosplit'] = self.__dosplit( self.__globals_( ) )
        _invocation_parameters['outputvis'] = self.__outputvis( self.__globals_( ) )
        _invocation_parameters['doconcat'] = self.__doconcat( self.__globals_( ) )
        _invocation_parameters['concatvis'] = self.__concatvis( self.__globals_( ) )
        _invocation_parameters['keep_orig_ms'] = self.__keep_orig_ms( self.__globals_( ) )

        try:
            with open(_postfile,'w') as _f:
                for _i in _invocation_parameters:
                    _f.write("%-12s = %s\n" % (_i,noobj(repr(_invocation_parameters[_i]))))
                _f.write("#calibeovsa( ")
                count = 0
                for _i in _invocation_parameters:
                    _f.write("%s=%s" % (_i,noobj(repr(_invocation_parameters[_i]))))
                    count += 1
                    if count < len(_invocation_parameters): _f.write(",")
                _f.write(" )\n")
        except: return False
        return True

    def __call__( self, vis=None, caltype=None, caltbdir=None, interp=None, docalib=None, doflag=None, flagant=None, doimage=None, imagedir=None, antenna=None, timerange=None, spw=None, stokes=None, dosplit=None, outputvis=None, doconcat=None, concatvis=None, keep_orig_ms=None ):
        def noobj(s):
           if s.startswith('<') and s.endswith('>'):
               return "None"
           else:
               return s
        _prefile = os.path.realpath('calibeovsa.pre')
        _postfile = os.path.realpath('calibeovsa.last')
        _return_result_ = None
        _arguments = [vis,caltype,caltbdir,interp,docalib,doflag,flagant,doimage,imagedir,antenna,timerange,spw,stokes,dosplit,outputvis,doconcat,concatvis,keep_orig_ms]
        _invocation_parameters = OrderedDict( )
        if any(map(lambda x: x is not None,_arguments)):
            # invoke python style
            # set the non sub-parameters that are not None
            local_global = { }
            if vis is not None: local_global['vis'] = vis
            if caltype is not None: local_global['caltype'] = caltype
            if caltbdir is not None: local_global['caltbdir'] = caltbdir
            if interp is not None: local_global['interp'] = interp
            if docalib is not None: local_global['docalib'] = docalib
            if doflag is not None: local_global['doflag'] = doflag
            if doimage is not None: local_global['doimage'] = doimage
            if dosplit is not None: local_global['dosplit'] = dosplit
            if doconcat is not None: local_global['doconcat'] = doconcat

            # the invocation parameters for the non-subparameters can now be set - this picks up those defaults
            _invocation_parameters['vis'] = self.__vis( local_global )
            _invocation_parameters['caltype'] = self.__caltype( local_global )
            _invocation_parameters['caltbdir'] = self.__caltbdir( local_global )
            _invocation_parameters['interp'] = self.__interp( local_global )
            _invocation_parameters['docalib'] = self.__docalib( local_global )
            _invocation_parameters['doflag'] = self.__doflag( local_global )
            _invocation_parameters['doimage'] = self.__doimage( local_global )
            _invocation_parameters['dosplit'] = self.__dosplit( local_global )
            _invocation_parameters['doconcat'] = self.__doconcat( local_global )

            # the sub-parameters can then be set. Use the supplied value if not None, else the function, which gets the appropriate default
            _invocation_parameters['flagant'] = self.__flagant( _invocation_parameters ) if flagant is None else flagant
            _invocation_parameters['imagedir'] = self.__imagedir( _invocation_parameters ) if imagedir is None else imagedir
            _invocation_parameters['antenna'] = self.__antenna( _invocation_parameters ) if antenna is None else antenna
            _invocation_parameters['timerange'] = self.__timerange( _invocation_parameters ) if timerange is None else timerange
            _invocation_parameters['spw'] = self.__spw( _invocation_parameters ) if spw is None else spw
            _invocation_parameters['stokes'] = self.__stokes( _invocation_parameters ) if stokes is None else stokes
            _invocation_parameters['outputvis'] = self.__outputvis( _invocation_parameters ) if outputvis is None else outputvis
            _invocation_parameters['concatvis'] = self.__concatvis( _invocation_parameters ) if concatvis is None else concatvis
            _invocation_parameters['keep_orig_ms'] = self.__keep_orig_ms( _invocation_parameters ) if keep_orig_ms is None else keep_orig_ms

        else:
            # invoke with inp/go semantics
            _invocation_parameters['vis'] = self.__vis( self.__globals_( ) )
            _invocation_parameters['caltype'] = self.__caltype( self.__globals_( ) )
            _invocation_parameters['caltbdir'] = self.__caltbdir( self.__globals_( ) )
            _invocation_parameters['interp'] = self.__interp( self.__globals_( ) )
            _invocation_parameters['docalib'] = self.__docalib( self.__globals_( ) )
            _invocation_parameters['doflag'] = self.__doflag( self.__globals_( ) )
            _invocation_parameters['flagant'] = self.__flagant( self.__globals_( ) )
            _invocation_parameters['doimage'] = self.__doimage( self.__globals_( ) )
            _invocation_parameters['imagedir'] = self.__imagedir( self.__globals_( ) )
            _invocation_parameters['antenna'] = self.__antenna( self.__globals_( ) )
            _invocation_parameters['timerange'] = self.__timerange( self.__globals_( ) )
            _invocation_parameters['spw'] = self.__spw( self.__globals_( ) )
            _invocation_parameters['stokes'] = self.__stokes( self.__globals_( ) )
            _invocation_parameters['dosplit'] = self.__dosplit( self.__globals_( ) )
            _invocation_parameters['outputvis'] = self.__outputvis( self.__globals_( ) )
            _invocation_parameters['doconcat'] = self.__doconcat( self.__globals_( ) )
            _invocation_parameters['concatvis'] = self.__concatvis( self.__globals_( ) )
            _invocation_parameters['keep_orig_ms'] = self.__keep_orig_ms( self.__globals_( ) )
        try:
            with open(_prefile,'w') as _f:
                for _i in _invocation_parameters:
                    _f.write("%-12s = %s\n" % (_i,noobj(repr(_invocation_parameters[_i]))))
                _f.write("#calibeovsa( ")
                count = 0
                for _i in _invocation_parameters:
                    _f.write("%s=%s" % (_i,noobj(repr(_invocation_parameters[_i]))))
                    count += 1
                    if count < len(_invocation_parameters): _f.write(",")
                _f.write(" )\n")
        except: pass
        try:
            _return_result_ = _calibeovsa_t( _invocation_parameters['vis'],_invocation_parameters['caltype'],_invocation_parameters['caltbdir'],_invocation_parameters['interp'],_invocation_parameters['docalib'],_invocation_parameters['doflag'],_invocation_parameters['flagant'],_invocation_parameters['doimage'],_invocation_parameters['imagedir'],_invocation_parameters['antenna'],_invocation_parameters['timerange'],_invocation_parameters['spw'],_invocation_parameters['stokes'],_invocation_parameters['dosplit'],_invocation_parameters['outputvis'],_invocation_parameters['doconcat'],_invocation_parameters['concatvis'],_invocation_parameters['keep_orig_ms'] )
        except Exception as e:
            from traceback import format_exc
            from casatasks import casalog
            casalog.origin('calibeovsa')
            casalog.post("Exception Reported: Error in calibeovsa: %s" % str(e),'SEVERE')
            casalog.post(format_exc( ))
            _return_result_ = False
        try:
            os.rename(_prefile,_postfile)
        except: pass
        return _return_result_

calibeovsa = _calibeovsa( )




import lmfit
import numpy as np
from suncasa.pygsfit import gsutils
from importlib import reload

reload(gsutils)

# msfile = 'IDB20210507_1840-1920XXYY.cal.10s.slfcaled.ms'
## SDO/AIA image fits
aiafile = 'aia.lev1_euv_12s.2021-05-07T190209Z.171.image_lev1.fits'
# aiafile = 'aia.lev1_uv_24s.2021-05-07T190214Z.1600.image_lev1.fits'
## EOVSA image fits created above
eofiles = ['EOVSA_20210507T190205.000000.outim.image.allbd.fits']
# eofiles = ['EOVSA_20210507T190135.000000.outim.image.allbd.fits','EOVSA_20210507T190205.000000.outim.image.allbd.fits']
spws = ['{}'.format(l) for l in range(48)]
## image center ("xycen"), pixel scale ("cell"), and image field of view ("fov")
## for the plots
xycen = np.array([-895, 290])
fov = np.array([200, 200])
## The alpha blending value for , between 0 (transparent) and 1 (opaque).
calpha = 0.5
## The respective maximum intensity of EOVSA images
clevels = np.array([0.3, 1.0])
freqghz_bound = [2., 14.0]
# freqghz_bound = [-1, 100]

gt = gsutils.GStool(eofiles, aiafile=aiafile, xycen=xycen, fov=fov, freqghz_bound=freqghz_bound,
                    calpha=calpha, clevels=clevels)

dep = 20.  # total source depth, 1e8 cm
Bmag = 30.
Bmagerr = 0.0
delta = 4.0
deltaerr = 0.0
# em = region.em
# nth = np.sqrt(em / (dep * 725.e5)) / 1e10
nth = 1.0
ntherr = 0.0
nrl = 1.0e6
nrlerr = 0.0
E_hi = 0.1
Emin = 0.1
Emax = 5.
nrlh = nrl * (E_hi ** (1. - delta) - Emax ** (1. - delta)) / (Emin ** (1. - delta) - Emax ** (1. - delta)) + 1e-8
nrlherr = 0.0
lognrlh = np.log10(nrlh)
lognrlherr = nrlherr / nrlh / np.log(10.)
Tth = 30.  # Thermal temperature in MK
# Tth = region.wT / 1e6  # Thermal temperature in MK
theta = 70.

params = lmfit.Parameters()
params.add('lnf', value=-2., vary=False)  # ln of fractional error adjustment

# params.add('ssz2', value=ssz2, vary=False)  # pixel size in arcsec
params.add('depth', value=dep, min=5.0, max=300., vary=False)  # Column depth, 1e8 cm (1 Mm)
params.add('Bmag', value=Bmag, min=20., max=100., vary=True)  # Magnetic field strength in G
params.add('Tth', value=Tth, min=5, max=30., vary=False)  # Thermal temperature in MK
params.add('nth', value=nth, min=0.1, max=20.0, vary=False)  # Thermal density in 1e10 cm^{-3}
params.add('lognrlh', value=np.log10(nrlh), min=1., max=10., vary=True)  # Non-thermal density in cm^{-3}
params.add('delta', value=delta, min=1.5, max=10.0, vary=True)  # Power law index for nonthermal electrons
params.add('theta', value=theta, min=20., max=85, vary=False)  # theta
params.add('Emin', value=Emin, min=0.001, max=0.05, vary=False)  # Emin
params.add('Emax', value=Emax, min=0.1, max=300, vary=False)  # Emax

gt.set_params(params=params)

gt.fit()



import numpy as np
import numpy.ma as ma
import os, struct
from time import time
import multiprocessing as mprocs
from suncasa.utils import DButil

from ...casa_compat import import_casatools, import_casatasks
tasks = import_casatasks('casalog')
casalog = tasks.get('casalog')

tools = import_casatools(['iatool', 'rgtool'])
iatool = tools['iatool']
rgtool = tools['rgtool']

def maxfit_iter(imgfiles, box, width, imidx):
    myia = iatool()
    myrg = rgtool()
    try:
        from astropy.io import fits as pyfits
    except:
        try:
            import pyfits
        except ImportError:
            raise ImportError('Neither astropy nor pyfits exists in this CASA installation')
    img = imgfiles[imidx]

    try:
        if (not myia.open(img)):
            raise Exception("Cannot create image analysis tool using " + img)
        print('Processing image: ' + img)
        hdulist = pyfits.open(img)
        hdu = hdulist[0]
        hdr = pyfits.getheader(img)
        pols = DButil.polsfromfitsheader(hdr)
        freqs = DButil.freqsfromfitsheader(hdr)
        ndx, ndy, nchans, npols = myia.shape()
        blc, trc = [0, 0], [ndx, ndy]
        if 'box' in locals():
            if box != '':
                blc[0], blc[1], trc[0], trc[1] = [int(ll) for ll in box.split(',')]
        if 'width' not in locals():
            width = 5
        results = {}
        for itpp in pols:
            results[itpp] = {'results': {}, 'converged': []}
        for ll in range(nchans):
            for pp, itpp in enumerate(pols):
                imgdata = hdu.data[pp, ll, blc[1]:trc[1] + 1, blc[0]:trc[0] + 1]
                imgdata = np.ma.masked_less(imgdata, 0.5 * np.nanmax(imgdata))
                ny, nx = imgdata.shape
                XX, YY = np.meshgrid(np.arange(nx), np.arange(ny))
                imgdata_avg = imgdata.mean()
                wx, wy = (XX * imgdata).mean() / imgdata_avg, (YY * imgdata).mean() / imgdata_avg
                comp = 'component{}'.format(ll)
                r = myrg.box(blc=[blc[0], blc[1], ll, pp], trc=[trc[0], trc[1], ll, pp])
                iachan = myia.subimage(region=r, dropdeg=True)
                try:
                    result_dict = iachan.maxfit(point=True, negfind=False, width = width)
                    result_dict['component0']['centroid'] = iachan.toworld([wx, wy], 'm')['measure']
                    result_dict['component0']['converged'] = True
                    result_dict['component0']['flux']['polarisation'] = itpp
                    result_dict['component0']['spectrum']['frequency']['m0']['value'] = float(freqs[ll])
                    results[itpp]['results'][comp] = result_dict['component0']
                    results[itpp]['converged'].append(True)
                except:
                    results[itpp]['converged'].append(False)
        results[itpp]['results']['nelements'] = results[itpp]['results'].keys()
        # update timestamp
        timstr = hdr['date-obs']
        return [True, timstr, img, results]
    except Exception as instance:
        casalog.post(str('*** Error in imfit ***') + str(instance))
        # raise instance
        return [False, timstr, img, {}]
    finally:
        myia.done()


def pmaxfit(imagefiles, ncpu, box, width):
    from functools import partial
    # check if imagefiles is a single file or a list of files

    if isinstance(imagefiles, str):
        imagefiles = [imagefiles]
    if (not isinstance(imagefiles, list)):
        print('input "imagefiles" is not a list. Abort...')

    # check file existence
    imgfiles = []
    for i, img in enumerate(imagefiles):
        if os.path.exists(img):
            imgfiles.append(img)
        else:
            casalog.post(img + 'does not exist. Skipping this one...')

    iterable = range(len(imgfiles))
    res = []

    if not (type(ncpu) is int):
        casalog.post('ncpu should be an integer')
        ncpu = 8

    # partition
    maxfit_part = partial(maxfit_iter, imgfiles, box, width)

    # parallelization
    para = 1
    timelapse = 0
    t0 = time()
    if para:
        casalog.post('Perform maxfit in parallel ...')
        pool = mprocs.Pool(ncpu)
        # res = pool.map_async(maxfit_part, iterable)
        res = pool.map(maxfit_part, iterable)
        pool.close()
        pool.join()
    else:
        for i in iterable:
            res.append(maxfit_part(i))

    t1 = time()
    timelapse = t1 - t0
    print('It took %f secs to complete' % timelapse)
    # repackage this into a single dictionary
    results = {'succeeded': [], 'timestamps': [], 'imagenames': [], 'outputs': []}
    for r in res:
        results['succeeded'].append(r[0])
        results['timestamps'].append(r[1])
        results['imagenames'].append(r[2])
        results['outputs'].append(r[3])

    return results



import argparse
import subprocess
import os
import glob
from tqdm import tqdm
import matplotlib.pyplot as plt


def create_movie_from_images(
        input_path,
        output_name='output_movie',
        image_format='png',
        frame_rate=10,
        resolution=None,
        quality=15,
        start_frame=0,
        crop_region=None,
        timestamps=None,
        image_dpi=200,
        preserve_temp=False,
        use_existing_temp=False,
        allow_rotation=True,
        force_overwrite=False
):
    """
    Create a movie from a series of images.

    :param str input_path: Path to image files or list of image paths
    :param str output_name: Name of the output movie file (without extension)
    :param str image_format: Format of the input image files
    :param int frame_rate: Frames per second in the output movie
    :param str resolution: Output movie resolution (format: 'widthxheight')
    :param int quality: Video quality (0-51, lower is better)
    :param int start_frame: Starting number for image sequence
    :param list crop_region: Crop images: [left, right, top, bottom]
    :param list timestamps: List of timestamps for each frame
    :param int image_dpi: DPI for image processing
    :param bool preserve_temp: Keep temporary files after processing
    :param bool use_existing_temp: Use existing images in the temp folder
    :param bool allow_rotation: Allow automatic image rotation
    :param bool force_overwrite: Overwrite existing output file
    :return: None
    """
    if isinstance(input_path, list):
        image_files = input_path
        base_dir = os.path.dirname(input_path[0])
    else:
        image_files = sorted(glob.glob(f"{input_path}*.{image_format}"))
        base_dir = os.path.dirname(input_path)

    if not image_files:
        print('No images found!')
        return

    temp_dir = os.path.join(base_dir, 'temp_images') + os.path.sep
    os.makedirs(temp_dir, exist_ok=True)

    if not use_existing_temp:
        process_images(image_files, temp_dir, image_format, crop_region, timestamps, resolution, image_dpi)

    ffmpeg_options = {
        'r': frame_rate,
        's': resolution,
        'start_number': start_frame,
        'crf': quality
    }
    if resolution is None:
        ffmpeg_options.pop('s')

    overwrite_flag = '-y' if force_overwrite else '-n'
    rotation_flag = '' if allow_rotation else '-noautorotate'

    ffmpeg_args = ' '.join([f'-{k} {v}' for k, v in ffmpeg_options.items()])

    output_path = f"{os.path.join(base_dir, output_name)}"
    if not output_path.lower().endswith('.mp4'):
        output_path += '.mp4'

    if not force_overwrite and os.path.exists(output_path):
        print(f"Error: Output file '{output_path}' already exists. Use -y or --force-overwrite to overwrite.")
        return

    overwrite_flag = '-y' if force_overwrite else '-n'
    rotation_flag = '' if allow_rotation else '-noautorotate'

    ffmpeg_command = (
        f'ffmpeg {rotation_flag} {overwrite_flag} -r {frame_rate} -f image2 '
        f'-i {temp_dir}%04d.{image_format} '
        f'-vcodec libx264 -pix_fmt yuv420p {ffmpeg_args} '
        f'-vf "pad=ceil(iw/2)*2:ceil(ih/2)*2" '
        f'{output_path}'
    )

    try:
        subprocess.check_output(['bash', '-c', ffmpeg_command], stderr=subprocess.STDOUT)
        print(f"Movie created: {output_path}")
    except subprocess.CalledProcessError as e:
        print(f"Error creating movie: {e.output.decode()}")

    print(f"FFmpeg command: {ffmpeg_command}")

    if not preserve_temp:
        os.system(f'rm -rf {temp_dir}')


def process_images(image_files, temp_dir, image_format, crop_region, timestamps, resolution, image_dpi):
    if crop_region or timestamps:
        plt.ioff()
        fig_size = [float(dim) / image_dpi for dim in resolution.split('x')] if resolution else None
        fig, ax = plt.subplots(figsize=fig_size)
        for idx, img_path in enumerate(tqdm(image_files, desc="Processing images")):
            data = plt.imread(img_path)
            if crop_region:
                left, right, top, bottom = crop_region
                data = data[top:bottom + 1, left:right + 1, :]
            if idx == 0:
                im = ax.imshow(data)
                ax.set_axis_off()
            else:
                im.set_array(data)
            if timestamps:
                ax.set_title(timestamps[idx])
            if idx == 0:
                fig.tight_layout()
            fig.savefig(f'{temp_dir}{idx:04d}.{image_format}', dpi=image_dpi, format=image_format)
        plt.close(fig)
    else:
        for idx, img_path in enumerate(tqdm(image_files, desc="Copying images")):
            os.system(f'cp {img_path} {temp_dir}{idx:04d}.{image_format}')


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Create a movie from a series of images.")
    parser.add_argument("input_path", help="Path to image files or list of image paths")
    parser.add_argument("-o", "--output-name", default="output_movie",
                        help="Name of the output movie file (without extension)")
    parser.add_argument("-f", "--image-format", default="png", help="Format of the input image files")
    parser.add_argument("-r", "--frame-rate", type=int, default=10, help="Frames per second in the output movie")
    parser.add_argument("-s", "--resolution", help="Output movie resolution (format: 'widthxheight')")
    parser.add_argument("-q", "--quality", type=int, default=15, help="Video quality (0-51, lower is better)")
    parser.add_argument("-n", "--start-frame", type=int, default=0, help="Starting number for image sequence")
    parser.add_argument("-c", "--crop-region", nargs=4, type=int, help="Crop images: left right top bottom")
    parser.add_argument("-t", "--timestamps", nargs="+", help="List of timestamps for each frame")
    parser.add_argument("-d", "--image-dpi", type=int, default=200, help="DPI for image processing")
    parser.add_argument("-p", "--preserve-temp", action="store_true", help="Keep temporary files after processing")
    parser.add_argument("-u", "--use-existing-temp", action="store_true", help="Use existing images in the temp folder")
    parser.add_argument("--no-rotation", dest="allow_rotation", action="store_false",
                        help="Disable automatic image rotation")
    parser.add_argument("-y", "--force-overwrite", action="store_true", help="Overwrite existing output file")

    args = parser.parse_args()

    create_movie_from_images(
        input_path=args.input_path,
        output_name=args.output_name,
        image_format=args.image_format,
        frame_rate=args.frame_rate,
        resolution=args.resolution,
        quality=args.quality,
        start_frame=args.start_frame,
        crop_region=args.crop_region,
        timestamps=args.timestamps,
        image_dpi=args.image_dpi,
        preserve_temp=args.preserve_temp,
        use_existing_temp=args.use_existing_temp,
        allow_rotation=args.allow_rotation,
        force_overwrite=args.force_overwrite
    )


import os
import sys
import time
from datetime import datetime, timedelta

import astropy.units as u
from matplotlib.dates import AutoDateFormatter, AutoDateLocator


# Function to format the y-axis as integer frequencies
def int_formatter(x, pos):
    return f"{int(x)}"


def setup_time_axis(ax, start, end, minticks=5, maxticks=10):
    """
    Set up the time axis for a plot with specific tick intervals based on the time range,
    ensuring a minimum and maximum number of ticks.

    :param ax: The axis to set up the time axis for.
    :type ax: matplotlib.axes.Axes
    :param start: The start time of the plot.
    :type start: astropy.time.Time
    :param end: The end time of the plot.
    :type end: astropy.time.Time
    :param minticks: Minimum number of major ticks desired.
    :type minticks: int
    :param maxticks: Maximum number of major ticks desired.
    :type maxticks: int
    """
    # Calculate the total time range in minutes
    total_minutes = (end - start).to(u.min).value

    # Create locator for major ticks
    major_locator = AutoDateLocator(minticks=minticks, maxticks=maxticks)
    ax.xaxis.set_major_locator(major_locator)

    # Create formatter for major ticks
    formatter = AutoDateFormatter(major_locator)
    if total_minutes < 1:
        formatter.scaled[1.0] = '%H:%M:%S.%f'  # Show milliseconds for very short durations
    elif total_minutes < 60:
        formatter.scaled[1.0] = '%H:%M:%S'  # Show seconds for short durations
    else:
        formatter.scaled[1.0] = '%H:%M'  # For longer durations, show only hours and minutes
    formatter.scaled[1 / 24] = '%H:%M'
    formatter.scaled[1 / (24 * 60)] = '%H:%M'
    formatter.scaled[1 / (24 * 60 * 60)] = '%H:%M:%S'
    ax.xaxis.set_major_formatter(formatter)

    # Create locator for minor ticks
    minor_locator = AutoDateLocator(minticks=minticks * 5, maxticks=minticks * 5)  # More minor ticks
    ax.xaxis.set_minor_locator(minor_locator)

    # Set x-axis limits
    ax.set_xlim(start.plot_date, end.plot_date)


def plot(timestamp=None, timerange=None, figdir='/common/lwa/spec_v2/daily/', figname=None, combine=True,
         clip=[10, 99.995], add_logo=False, fast_plot=True, interactive=False):
    """
    Plot the OVRO-LWA and EOVSA spectrograms along with STIX and GOES light curves for a given timestamp or time range.

    :param timestamp: The timestamp for the data to be plotted. If not provided, the current UTC time is used.
    :type timestamp: datetime.datetime, optional
    :param timerange: The time range to be plotted. If provided, it overrides the timestamp.
    :type timerange: list of datetime.datetime, optional
    :param figdir: Directory where the figures will be saved, defaults to '/common/lwa/spec_v2/daily/'.
    :type figdir: str, optional
    :param figname: The file name for the combined figure. If not provided, it is generated based on the timestamp.
    :type figname: str, optional
    :param combine: If True, combine all plots into a single figure, defaults to True. Otherwise, save each plot (OVRO-LWA, EOVSA, STIX, GOES) separately.
    :type combine: bool, optional
    :param clip: The percentile values for clipping the color scale of EOVSA and OVRO-LWA spectrograms, defaults to [10, 99.995].
    :type clip: list of float, optional
    :param add_logo: If True, add logos to the plots, defaults to False.
    :type add_logo: bool, optional
    :param fast_plot: If True, use fast plotting methods, defaults to True. If you want to plot the full resolution, set it to False. But it will take longer to plot. If the time range is less than 30 minutes, it will automatically set to False.
    :type fast_plot: bool, optional
    :param interactive: If False, suppress plot display and use the 'Agg' backend, defaults to False.
    :type interactive: bool, optional
    :return: List of file paths to the saved figures.
    :rtype: list of str

    Examples:
    ---------
    from suncasa.utils import ovsas_spectrogram as ovsp
    from datetime import datetime
    # Example 1: Plotting the synoptic spectrogram for 2024 July 31
    ovsp.plot(datetime(2024, 7, 31), figdir='/data1/workdir/')

    # Example 2: Plotting the synoptic spectrogram for a specific time interval on 2024 July 31
    ovsp.plot(timerange=[datetime(2024, 7, 31, 17, 20), datetime(2024, 7, 31, 20, 40)],
        figdir='/data1/workdir/', fast_plot=True, clip=[5, 99.995])

    # Example 3: Plotting the synoptic spectrogram for a specific time interval on 2024 July 31  in full resolution
    ovsp.plot(timerange=[datetime(2024, 7, 31, 18, 20), datetime(2024, 7, 31, 18, 40)],
        figdir='/data1/workdir/', fast_plot=False, clip=[5, 99.995])
    """
    t0 = time.time()

    import numpy as np
    import pandas as pd
    import sunpy.timeseries as ts
    from astropy.time import Time
    from stixdcpy.quicklook import LightCurves
    from sunpy.net import Fido
    from sunpy.net import attrs as a
    from sunpy.time import parse_time
    from suncasa.dspec import dspec as ds
    import matplotlib
    import matplotlib.colors as mcolors
    from matplotlib import pyplot as plt
    from mpl_toolkits.axes_grid1 import make_axes_locatable
    from matplotlib.ticker import FuncFormatter
    if add_logo:
        from ovrolwasolar.visualization import njit_logo_str, nsf_logo
        import base64
        import io
        import matplotlib.image as mpimg

    if not interactive:
        try:
            matplotlib.use('Agg')  # Use the Agg backend
        except Exception as e:
            print(f'Error: {e}')
            print('Failed to set the backend to Agg. Use default backend.')
        plt.ioff()
    else:
        try:
            if os.name == 'posix':
                if sys.platform == 'darwin':
                    matplotlib.use('MacOSX')
                else:
                    matplotlib.use('QtAgg')
        except Exception as e:
            print(f'Error: {e}')
            print('Failed to set the specified backend. Using default backend.')
        plt.ion()

    if timestamp is None:
        if timerange is not None:
            timestamp = timerange[0]
            if timerange[1] - timerange[0] < timedelta(minutes=30):
                fast_plot = False
        else:
            # Set the current timestamp if none is provided
            timestamp = datetime.utcnow() - timedelta(days=1)

    # Ensure the output directory exists
    os.makedirs(figdir, exist_ok=True)

    if combine:
        if figname is None:
            # Define the file name for the combined figure
            figname = os.path.join(figdir, f'fig-OVSAs_spec_{timestamp.strftime("%Y%m%d")}.png')
        if os.path.exists(figname):
            if timerange is None:
                # If the combined figure already exists, skip plotting individual figures
                print(f'Combined figure for {timestamp.strftime("%Y-%m-%d")} already exists. Skipping individual figures.')
                return [figname]
    else:
        # Define file names for each figure
        figname_eovsa = os.path.join(figdir, f'fig-eovsa_spec_{timestamp.strftime("%Y%m%d")}.png')
        figname_ovrolwa = os.path.join(figdir, f'fig-ovrolwa_spec_{timestamp.strftime("%Y%m%d")}.png')
        figname_stix = os.path.join(figdir, f'fig-stix_lc_{timestamp.strftime("%Y%m%d")}.png')
        figname_goes = os.path.join(figdir, f'fig-goes_lc_{timestamp.strftime("%Y%m%d")}.png')

    # Define paths for the data files
    ovrolwapath = '/common/lwa/spec_v2/fits/'
    ovrolwa_specfile = os.path.join(ovrolwapath, timestamp.strftime("%Y%m%d.fits"))
    eovsapath = f'/data1/eovsa/fits/synoptic/{timestamp.strftime("%Y/%m/%d")}/'
    eovsa_specfile = os.path.join(eovsapath, timestamp.strftime("EOVSA_TPall_%Y%m%d.fts"))

    # Define default time range if no data is available
    default_start_time = Time(timestamp.replace(hour=14, minute=0, second=0))
    default_end_time = Time((timestamp + timedelta(days=1)).replace(hour=2, minute=0, second=0))

    if combine:
        fig, axs = plt.subplots(4, 1, figsize=(12, 12), sharex=True, height_ratios=[2, 2, 3, 4])
        ax_goes, ax_stix, ax_eovsa, ax_ovrolwa = axs
    else:
        # Initialize figures and axes
        fig_ovrolwa, ax_ovrolwa = plt.subplots(1, 1, figsize=(15, 4))
        fig_eovsa, ax_eovsa = plt.subplots(1, 1, figsize=(15, 4))
        fig_stix, ax_stix = plt.subplots(1, 1, figsize=(15, 3))
        fig_goes, ax_goes = plt.subplots(1, 1, figsize=(15, 3))

    # Load OVRO-LWA spectrogram data
    print(f'processing OVRO-LWA spectrogram data for {timestamp.strftime("%Y-%m-%d")}')
    if os.path.exists(ovrolwa_specfile):
        cmap = 'viridis'
        ovrolwa_bkgtim = None
        d_ovrolwa = ds.Dspec()
        d_ovrolwa.read(ovrolwa_specfile, source='lwa')
        d_ovrolwa.pol = 'IV'
        timerange_ovrolwa = None
        if timerange is not None:
            timerange_ovrolwa = list(Time(timerange).iso)
        vmax = np.nanpercentile(d_ovrolwa.data, 99.995)
        vmin = np.nanpercentile(d_ovrolwa.data, 5)
        norm_I_ovrolwa = mcolors.LogNorm(vmin=vmin, vmax=vmax)
        d_ovrolwa.plot(pol='I', timerange=timerange_ovrolwa, bkgtim=ovrolwa_bkgtim, plot_fast=fast_plot,
                       norm=norm_I_ovrolwa,
                       percentile=clip, minmaxpercentile=True, freq_unit='MHz', cmap=cmap, axes=ax_ovrolwa)
        ovrolwa_tim = d_ovrolwa.time_axis
        ovro_lwa_start, ovro_lwa_end = ovrolwa_tim[0], ovrolwa_tim[-1]
    else:
        ax_ovrolwa.plot([], [])
        ax_ovrolwa.text(0.5, 0.5, 'No OVRO-LWA data available', transform=ax_ovrolwa.transAxes,
                        ha='center', va='center', fontsize=12, color='gray')
        ax_ovrolwa.set_ylabel('Frequency [MHz]')
        ax_ovrolwa.set_ylim(29.033934, 83.871824)
        ovro_lwa_start, ovro_lwa_end = default_start_time, default_end_time
        divider = make_axes_locatable(ax_ovrolwa)
        cax_spec = divider.append_axes('right', size='1.5%', pad=0.05)
        cax_spec.set_visible(False)

    ax_ovrolwa.set_yscale('log')
    ax_ovrolwa.yaxis.set_minor_formatter(FuncFormatter(int_formatter))
    ax_ovrolwa.set_title(f'OVRO-LWA Dynamic Spectrum (Stokes I)')

    print(f'processing EOVSA spectrogram data for {timestamp.strftime("%Y-%m-%d")}')
    # Load EOVSA spectrogram data
    ax = ax_eovsa
    if os.path.exists(eovsa_specfile):
        cmap = 'viridis'
        eovsa_bkgtim = None
        timerange_eovsa = None
        if timerange is not None:
            timerange_eovsa = list(Time(timerange).iso)

        # norm_I_eovsa = mcolors.Normalize(vmin=None, vmax=None)
        d_eovsa = ds.Dspec()
        d_eovsa.read(eovsa_specfile, source='eovsa')
        data = d_eovsa.data
        bkgspec = np.nanpercentile(data, 1, axis=1)
        d_eovsa.data = data - bkgspec[:, np.newaxis]
        d_eovsa.data[d_eovsa.data < 0] = 0
        fghz_eovsa = d_eovsa.freq_axis/1e9
        bad_fghz = [1.742, 2.72]
        d_eovsa.data[np.logical_and(fghz_eovsa > bad_fghz[0], fghz_eovsa < bad_fghz[1])] = np.nan
        vmax = np.nanpercentile(d_eovsa.data, 99.995)
        vmin = np.nanpercentile(d_eovsa.data, 5)
        norm_I_eovsa = mcolors.LogNorm(vmin=vmin, vmax=vmax)
        d_eovsa.plot(pol='I', timerange=timerange_eovsa, bkgtim=eovsa_bkgtim, plot_fast=False, norm=norm_I_eovsa,
                     percentile=clip, minmaxpercentile=True, freq_unit='GHz', cmap=cmap, axes=ax_eovsa)
        eovsa_tim = d_eovsa.time_axis
        eovsa_start, eovsa_end = eovsa_tim[0], eovsa_tim[-1]
    else:
        ax_eovsa.plot([], [])
        ax_eovsa.text(0.5, 0.5, 'No EOVSA data available', transform=ax_eovsa.transAxes,
                      ha='center', va='center', fontsize=12, color='gray')
        ax_eovsa.set_ylabel('Frequency [GHz]')
        ax_eovsa.set_ylim(1.1053711, 17.979687)
        eovsa_start, eovsa_end = default_start_time, default_end_time
        divider = make_axes_locatable(ax_eovsa)
        cax_spec = divider.append_axes('right', size='1.5%', pad=0.05)
        cax_spec.set_visible(False)

    ax_eovsa.set_yscale('log')
    ax_eovsa.yaxis.set_major_formatter(FuncFormatter(int_formatter))
    ax_eovsa.yaxis.set_minor_formatter(FuncFormatter(int_formatter))
    ax_eovsa.set_title(f'EOVSA Dynamic Spectrum (Stokes I)')

    if timerange is None:
        # Determine the overall time range based on available data
        overall_start = min(ovro_lwa_start, eovsa_start)
        overall_end = max(ovro_lwa_end, eovsa_end)
    else:
        ## if timerange is provided, override the overall time range
        overall_start = Time(timerange[0])
        overall_end = Time(timerange[1])
        if combine:
            figname = os.path.join(figdir,
                                   f'fig-OVSAs_spec_{timerange[0].strftime("%Y%m%dT%H%M%S")}-{timerange[1].strftime("%Y%m%dT%H%M%S")}.png')
        else:
            figname_eovsa = os.path.join(figdir,
                                         f'fig-eovsa_spec_{timerange[0].strftime("%Y%m%dT%H%M%S")}-{timerange[1].strftime("%Y%m%dT%H%M%S")}.png')
            figname_ovrolwa = os.path.join(figdir,
                                           f'fig-ovrolwa_spec_{timerange[0].strftime("%Y%m%dT%H%M%S")}-{timerange[1].strftime("%Y%m%dT%H%M%S")}.png')
            figname_stix = os.path.join(figdir,
                                        f'fig-stix_lc_{timerange[0].strftime("%Y%m%dT%H%M%S")}-{timerange[1].strftime("%Y%m%dT%H%M%S")}.png')
            figname_goes = os.path.join(figdir,
                                        f'fig-goes_lc_{timerange[0].strftime("%Y%m%dT%H%M%S")}-{timerange[1].strftime("%Y%m%dT%H%M%S")}.png')

    print(f'processing STIX light curves for {timestamp.strftime("%Y-%m-%d")}')
    # Load STIX light curves
    lc = LightCurves.from_sdc(start_utc=overall_start.iso, end_utc=overall_end.iso, ltc=True)
    if len(lc.counts) > 0:
        lc.peek(ax=ax_stix)
    else:
        ax_stix.plot([], [])
        ax_stix.set_ylabel('Counts')
        ax_stix.set_ylim(20, 2e6)
        ax_stix.text(0.5, 0.5, 'No STIX data available', transform=ax_stix.transAxes,
                     ha='center', va='center', fontsize=12, color='gray')
        ax_stix.set_yscale('log')

    ax_stix.set_title(f'STIX Quick-look Light Curves')
    divider = make_axes_locatable(ax_stix)
    cax_spec = divider.append_axes('right', size='1.5%', pad=0.05)
    cax_spec.set_visible(False)
    if combine:
        pass
    else:
        fig_stix.subplots_adjust(bottom=0.15)

    print(f'processing GOES X-ray light curves for {timestamp.strftime("%Y-%m-%d")}')
    # Download GOES X-ray light curves
    try:
        try:
            goes_query = Fido.search(a.Time(overall_start.datetime, overall_end.datetime), a.Instrument('XRS'),
                                     a.Resolution("flx1s"), a.goes.SatelliteNumber(18))
        except Exception as e:
            print(f'Error: {e}. Loose search for GOES data')
            goes_query = Fido.search(a.Time(overall_start.datetime, overall_end.datetime), a.Instrument('XRS'))
        goes_files = Fido.fetch(goes_query)
        if goes_files:
            goes = ts.TimeSeries(goes_files)
            if isinstance(goes, list):
                df_comb = goes[0].to_dataframe()
                df_comb = df_comb[(df_comb["xrsa_quality"] == 0) & (df_comb["xrsb_quality"] == 0)]
                for g in goes[1:]:
                    df = g.to_dataframe()
                    df = df[(df["xrsa_quality"] == 0) & (df["xrsb_quality"] == 0)]
                    df_comb = pd.concat([df_comb, df])
            else:
                df_comb = goes.to_dataframe()
            units = dict([("xrsa", u.W / u.m ** 2), ("xrsb", u.W / u.m ** 2)])
            meta = dict({"instrument": "GOES X-ray sensor", "measurements": "primary", "type": "quicklook"})
            goes_ts = ts.TimeSeries(df_comb[['xrsa', 'xrsb']], meta, units, source="xrs")
        else:
            goes_json_data = pd.read_json("https://services.swpc.noaa.gov/json/goes/primary/xrays-7-day.json")
            # This will get us the short wavelength data.
            goes_short = goes_json_data[goes_json_data["energy"] == "0.05-0.4nm"]
            # This will get us the long wavelength data.
            goes_long = goes_json_data[goes_json_data["energy"] == "0.1-0.8nm"]

            time_array = parse_time(goes_short["time_tag"])
            filtered_indices = (time_array >= overall_start) & (time_array <= overall_end)
            filtered_short = goes_short[filtered_indices]
            filtered_long = goes_long[filtered_indices]

            # Create a DataFrame with the filtered data
            filtered_time_array = time_array[filtered_indices]
            goes_data = pd.DataFrame({
                "xrsa": filtered_short["flux"].values,
                "xrsb": filtered_long["flux"].values
            }, index=filtered_time_array.datetime)
            units = dict([("xrsa", u.W / u.m ** 2), ("xrsb", u.W / u.m ** 2)])
            meta = dict({"instrument": "GOES X-ray sensor", "measurements": "primary", "type": "quicklook"})
            goes_ts = ts.TimeSeries(goes_data, meta, units, source="xrs")
        goes_ts.plot(axes=ax_goes)
    except Exception as e:
        print(f'Error: {e}. Proceeding without GOES data.')
        ax_goes.plot([], [])
        ax_goes.set_ylim(1e-9, 1e-2)
        ax_goes.text(0.5, 0.5, 'No GOES data available', transform=ax_goes.transAxes,
                     ha='center', va='center', fontsize=12, color='gray')
        ax_goes.set_yscale('log')
    ax_goes.set_ylabel('Flux [W/m²]')

    divider = make_axes_locatable(ax_goes)
    cax_spec = divider.append_axes('right', size='1.5%', pad=0.05)
    cax_spec.set_visible(False)
    ax_goes.set_title(f'GOES X-ray Light Curves')
    if combine:
        pass
    else:
        fig_goes.subplots_adjust(bottom=0.15)

    # Set up the time axis for each plot
    setup_time_axis(ax_ovrolwa, overall_start, overall_end)
    setup_time_axis(ax_eovsa, overall_start, overall_end)
    setup_time_axis(ax_stix, overall_start, overall_end)
    setup_time_axis(ax_goes, overall_start, overall_end)

    for ax in [ax_ovrolwa, ax_eovsa, ax_stix, ax_goes]:
        ax.set_xlabel(f'Time [UTC, {timestamp.strftime("%Y %b %d")}]')

    if combine:
        for ax in axs[:-1]:
            ax.tick_params(axis='x', which='both', labelbottom=False)
            ax.set_xlabel('')
        fig.tight_layout()
        if add_logo:
            ax_logo1 = fig.add_axes([0.89, 0.91, 0.15, 0.08])
            img1 = base64.b64decode(njit_logo_str)
            img1 = io.BytesIO(img1)
            img1 = mpimg.imread(img1, format='png')
            ax_logo1.imshow(img1)
            ax_logo1.axis('off')

            ax_logo2 = fig.add_axes([0.81, 0.91, 0.15, 0.09])
            img2 = base64.b64decode(nsf_logo)
            img2 = io.BytesIO(img2)
            img2 = mpimg.imread(img2, format='png')
            ax_logo2.imshow(img2)
            ax_logo2.axis('off')
        fig.savefig(figname, dpi=300)
        print(f'Saved combined figure to {figname}')
        figout = [figname]
    else:
        if add_logo:
            for fig in [fig_ovrolwa, fig_eovsa, fig_stix, fig_goes]:
                ax_logo1 = fig.add_axes([0.89, 0.91, 0.15, 0.08])
                img1 = base64.b64decode(njit_logo_str)
                img1 = io.BytesIO(img1)
                img1 = mpimg.imread(img1, format='png')
                ax_logo1.imshow(img1)
                ax_logo1.axis('off')

                ax_logo2 = fig.add_axes([0.81, 0.91, 0.15, 0.09])
                img2 = base64.b64decode(nsf_logo)
                img2 = io.BytesIO(img2)
                img2 = mpimg.imread(img2, format='png')
                ax_logo2.imshow(img2)
                ax_logo2.axis('off')
        fig_eovsa.savefig(figname_eovsa, dpi=300)
        fig_ovrolwa.savefig(figname_ovrolwa, dpi=300)
        fig_stix.savefig(figname_stix, dpi=300)
        fig_goes.savefig(figname_goes, dpi=300)
        print(f'Saved EOVSA figure to {figname_eovsa}')
        print(f'Saved OVRO-LWA figure to {figname_ovrolwa}')
        print(f'Saved STIX figure to {figname_stix}')
        print(f'Saved GOES figure to {figname_goes}')
        figout = [figname_eovsa, figname_ovrolwa, figname_stix, figname_goes]
    #
    # plt.show()  # Optionally show all figures
    if not interactive:
        plt.close('all')

    t1 = time.time()
    print(f'Plotting took {t1 - t0:.2f} seconds.')
    return figout



import copy
import os
import numpy as np
from astropy.io import fits
from sunpy import map as smap
import warnings
import sys

warnings.simplefilter("ignore")

stokesval = {'1': 'I', '2': 'Q', '3': 'U', '4': 'V', '-1': 'RR', '-2': 'LL', '-3': 'RL', '-4': 'LR', '-5': 'XX',
             '-6': 'YY', '-7': 'XY', '-8': 'YX'}


def is_compressed_fits(fitsfile):
    '''
    Function to check if the FITS file contains compressed data
    '''
    with fits.open(fitsfile) as hdul:
        for hdu in hdul:
            if isinstance(hdu, fits.hdu.CompImageHDU):
                return True
        return False


def headerfix(header, PC_coor=True):
    '''
	this code fixes the header problem of fits out from CASA 5.4+ which leads to a streched solar image.
    Setting PC_coor equal to True will reset the rotation matrix.
    '''

    keys2remove = []
    for k in header:
        if k.upper().startswith('PC'):
            if not k.upper().startswith('PC0'):
                pcidxs = k.upper().replace('PC', '')
                hd_ = 'PC0' + pcidxs
                keys2remove.append(k)
                if PC_coor:
                    pcidx0, pcidx1 = pcidxs.split('_')
                    if pcidx0 == pcidx1:
                        header[hd_] = 1.0
                    else:
                        header[hd_] = 0.0
                else:
                    header[hd_] = header[k]
    for k in keys2remove:
        header.remove(k)
    return header


def headerparse(header):
    '''
        get axis index of polarization
    '''

    ndim = header['NAXIS']
    stokesinfo = {'axis': None, 'headernew': {}}
    keys2analy = ['NAXIS', 'CTYPE', 'CRVAL', 'CDELT', 'CRPIX', 'CUNIT']
    for dim in range(1, ndim + 1):
        k = 'CTYPE{}'.format(dim)
        if header[k].startswith('STOKES'):
            stokesinfo['axis'] = dim
    if stokesinfo['axis'] is not None:
        dim = stokesinfo['axis']
        for k in keys2analy:
            k_ = '{}{}'.format(k, dim)
            stokesinfo['headernew']['{}{}'.format(k, 4)] = header[k_]
    return stokesinfo


def headersqueeze(header, data):
    """
    Squeezes single-dimensional entries from an n-dimensional FITS image data array and updates the FITS header accordingly.

    This function is useful for preparing image data for astropy fits compression, which only supports 1D, 2D, or 3D images. It removes
    any single-dimensional entries from the shape of the data array and updates the corresponding FITS header keys to reflect the
    new dimensions.

    :param header: FITS header object containing the metadata of the image.
    :type header: astropy.io.fits.Header
    :param data: n-dimensional image data array.
    :type data: numpy.ndarray

    :return: A tuple of the updated header object and the squeezed data array.
    :rtype: (astropy.io.fits.Header, numpy.ndarray)

    .. note::
        This function only updates the header keys related to dimensions, coordinate types, values, increments, reference pixels,
        and units. Any specific header keys related to coordinate transformations (e.g., PC matrix) for dimensions higher than
        the third are also updated if necessary. The function does not handle higher-order WCS transformations beyond simple axis
        permutations and squeezes.
    """
    dshape = data.shape
    ndim = data.ndim
    # Count single-dimensional entries in the data shape
    nsdim = np.count_nonzero(np.array(dshape) == 1)
    # Calculate the number of non-single-dimensional entries
    nonsdim = ndim - nsdim

    # If there are no single-dimensional entries, return the original header and data
    if nsdim == 0:
        return header, data
    else:
        # Define header keys that might need to be changed due to squeezing
        keys2chng = ['NAXIS', 'CTYPE', 'CRVAL', 'CDELT', 'CRPIX', 'CUNIT']  # ,'PC01_', 'PC02_', 'PC03_', 'PC04_']

        idx_nonsdim = 0
        for idx, dim in enumerate(dshape[::-1]):
            # if dim>1: continue
            if dim > 1:
                idx_nonsdim = idx_nonsdim + 1
            for k in keys2chng:
                k_ = f'{k}{idx + 1}'
                v = header[k_]
                header.remove(k_)
                if dim > 1:
                    k_new = f'{k}{idx_nonsdim}'
                    header[k_new] = v
                else:
                    if k == 'CTYPE' and v.startswith('STOKES'):
                        header['STOKES'] = header[f'CRVAL{idx + 1}']

        idx_nonsdim1 = 0
        # Update PC matrix keys for non-single-dimensional entries
        for idx1, dim1 in enumerate(dshape[::-1]):
            if dim1 > 1:
                idx_nonsdim1 = idx_nonsdim1 + 1
            idx_nonsdim2 = 0
            for idx2, dim2 in enumerate(dshape[::-1]):
                if dim2 > 1:
                    idx_nonsdim2 = idx_nonsdim2 + 1
                k_ = f'PC{idx1 + 1:02d}_{idx2 + 1}'
                if k_ in header.keys():
                    v = header[k_]
                    header.remove(k_)
                    if dim1 > 1 and dim2 > 1:
                        k_new = f'PC{idx_nonsdim1:02d}_{idx_nonsdim2}'
                        header[k_new] = v

        header['NAXIS'] = nonsdim
        data = np.squeeze(data)
        return header, data


def get_bdinfo(freq, bw):
    """
    get band information from center frequencies and band widths.

    Parameters
    ----------
    freq : array_like
        an array of the center frequencies of all frequency bands in Hz
    bw: array_like
        an array of the band widths of all frequency bands in Hz

    Returns
    -------
    fbounds : `dict`
        A dict of band information
    """
    fghz = freq / 1.e9
    bwghz = bw / 1.e9
    bounds_lo = fghz - bwghz / 2.0
    bounds_hi = fghz + bwghz / 2.0
    bounds_all = np.hstack([bounds_lo, bounds_hi[-1]])
    fbounds = {'cfreqs': fghz, 'cfreqs_all': fghz, 'bounds_lo': bounds_lo,
               'bounds_hi': bounds_hi, 'bounds_all': bounds_all}
    return fbounds


def read(filepath, hdus=None, verbose=False, **kwargs):
    """
    Read a fits file.

    Parameters
    ----------
    filepath : `str`
        The fits file to be read.
    hdus: `int` or iterable
        The HDU indexes to read from the file.
    verbose: `bool`
        if verbose

    Returns
    -------
    pairs : `list`
        A list of (data, header) tuples

    Notes
    -----
    This routine reads all the HDU's in a fits file and returns a list of the
    data and a FileHeader instance for each one.

    Also all comments in the original file are concatenated into a single
    "comment" key in the returned FileHeader.
    """
    import collections

    with fits.open(filepath, ignore_blank=True) as hdulist:
        if hdus is not None:
            if isinstance(hdus, int):
                hdulist = hdulist[hdus]
            elif isinstance(hdus, collections.Iterable):
                hdulist = [hdulist[i] for i in hdus]

        hdulist = fits.hdu.HDUList(hdulist)
        for h in hdulist:
            h.verify('silentfix+warn')

        meta = {}
        for i, hdu in enumerate(hdulist):
            try:
                ndim = hdu.data.ndim
                header = hdu.header
                slc = [slice(None)] * ndim
                freq_axis = None
                pol_axis = None
                _ = None
                npol = 1
                nfreq_fits = 1
                for idx in range(ndim):
                    v = header['CTYPE{}'.format(idx + 1)]
                    if v.startswith('FREQ'):
                        freq_axis = ndim - (idx + 1)
                        nfreq_fits = header['NAXIS{}'.format(idx + 1)]
                    if v.startswith('STOKES'):
                        pol_axis = ndim - (idx + 1)
                        npol = header['NAXIS{}'.format(idx + 1)]
                if freq_axis is not None:
                    slc[freq_axis] = slice(0, 1)
                    meta['ref_cfreqs'] = (hdu.header['CRVAL{}'.format(ndim - freq_axis)] + hdu.header[
                        'CDELT{}'.format(ndim - freq_axis)] * np.arange(
                        hdu.header['NAXIS{}'.format(ndim - freq_axis)]))
                    meta['ref_freqdelts'] = np.ones_like(meta['ref_cfreqs']) \
                                            * hdu.header['CDELT{}'.format(ndim - freq_axis)]
                else:
                    meta['ref_cfreqs'] = np.array([hdu.header['RESTFRQ']])

                if pol_axis is not None:
                    slc[pol_axis] = slice(0, 1)
                    meta['pol_idxs'] = (hdu.header['CRVAL{}'.format(ndim - pol_axis)] + hdu.header[
                        'CDELT{}'.format(ndim - pol_axis)] * np.arange(
                        hdu.header['NAXIS{}'.format(ndim - pol_axis)]))
                    meta['pol_names'] = [stokesval['{0:d}'.format(int(p))] for p in meta['pol_idxs']]
                if _ is not None:
                    slc[_] = slice(0, 1)
                hdu.data[np.isnan(hdu.data)] = 0.0
                rmap = smap.Map(np.squeeze(hdu.data[tuple(slc)]), hdu.header)
                data = hdu.data.copy()
                meta['header'] = hdu.header.copy()
                meta['refmap'] = rmap  # this is a sunpy map of the first slice
                meta['naxis'] = ndim
                meta['hgln_axis'] = ndim - 1  # solar X
                meta['nx'] = header['NAXIS1']
                meta['hglt_axis'] = ndim - 2  # solar Y
                meta['ny'] = header['NAXIS2']
                meta['freq_axis'] = freq_axis
                meta['nfreq'] = nfreq_fits
                meta['pol_axis'] = pol_axis
                meta['npol'] = npol
                break
            except Exception as e:
                if verbose:
                    if hasattr(e, 'message'):
                        print(e.message)
                    else:
                        print(e)
                    print('skipped HDU {}'.format(i))
                meta, data = {}, None

        # Check if an additional frequency axis exists. If so, they should be in the last hdu.
        if hasattr(hdulist[-1].data, 'cfreqs'):
            if verbose:
                print('FITS file contains an additional frequency axis. '
                      'Update the frequency information in cfreqs and cdelts. '
                      'Ignore the original version with equal spacings.')
            meta['ref_cfreqs'] = np.array(hdulist[-1].data['cfreqs'])
            meta['ref_freqdelts'] = np.array(hdulist[-1].data['cdelts'])

        if hasattr(hdulist[-1].data, 'bmaj'):
            meta['bmaj'] = np.array(hdulist[-1].data['bmaj'])
            meta['bmin'] = np.array(hdulist[-1].data['bmin'])
            meta['bpa'] = np.array(hdulist[-1].data['bpa'])

        if hasattr(hdulist[-1].data, 'cbmaj'):
            meta['bmaj'] = np.array(hdulist[-1].data['cbmaj'])
            meta['bmin'] = np.array(hdulist[-1].data['cbmin'])
            meta['bpa'] = np.array(hdulist[-1].data['cbpa'])

        if hasattr(hdulist[-1].data, 'refra_shift_x'):
            meta['refra_shift_x'] = np.array(hdulist[-1].data['refra_shift_x'])
            meta['refra_shift_y'] = np.array(hdulist[-1].data['refra_shift_y'])

        else:
            if verbose:
                print('FITS file does not have an additional frequency axis. '
                      'Use the original version with equal spacings.')
        hdulist.close()

        return meta, data


def write(fname, data, header, mask=None, fix_invalid=True, filled_value=0.0, overwrite=True, **kwargs):
    """
    Take a data header pair and write a compressed FITS file.
    Caveat: only 1D, 2D, or 3D images are currently supported by Astropy fits compression.
    To be compressed, the image data array (n-dimensional) must have
    at least n-3 single-dimensional entries.

    Parameters
    ----------
    fname : `str`
        File name, with extension.
    data : `numpy.ndarray`
        n-dimensional data array.
    header : `dict`
        A header dictionary.
    compression_type: `str`, optional
        Compression algorithm: one of 'RICE_1', 'RICE_ONE', 'PLIO_1', 'GZIP_1', 'GZIP_2', 'HCOMPRESS_1'
    hcomp_scale: `float`, optional
        HCOMPRESS scale parameter
    """

    dshape = data.shape
    dim = data.ndim
    if dim - np.count_nonzero(np.array(dshape) == 1) > 3:
        return 0
    else:
        if fix_invalid:
            data[np.isnan(data)] = filled_value
        if kwargs is {}:
            kwargs.update({'compression_type': 'RICE_1', 'quantize_level': 4.0})
        if isinstance(fname, str):
            fname = os.path.expanduser(fname)

        if os.path.exists(fname):
            if overwrite:
                os.system('rm -rf {}'.format(fname))
            else:
                print('File exists. Set overwrite=True to overwrite the file.')
                return 0
        header, data = headersqueeze(header, data)
        hdunew = fits.CompImageHDU(data=data, header=header, **kwargs)
        if mask is None:
            hdulnew = fits.HDUList([fits.PrimaryHDU(), hdunew])
        else:
            hdumask = fits.CompImageHDU(data=mask.astype(np.uint8), **kwargs)
            hdulnew = fits.HDUList([fits.PrimaryHDU(), hdunew, hdumask])
        hdulnew.writeto(fname, output_verify='fix')
        return 1


def header_to_xml(header):
    import xml.etree.ElementTree as ET

    from datetime import datetime
    dt = datetime.now()
    formatted_dt = dt.strftime("%a %b %d %H:%M:%S %Y")

    # create the file structure
    tree = ET.ElementTree()
    root = ET.Element('meta')

    # Add fits section
    elem = ET.Element('fits')
    for k, v in header.items():
        child = ET.Element(k)
        if isinstance(v, bool):
            v = int(v)
        child.text = str(v)
        elem.append(child)
    root.append(elem)

    # Add helioviewer section
    hv_comment = f"""
    JP2 file generated by {__file__} on {formatted_dt} at EOVSA (NJIT).
    For inquiries or feedback regarding this file, please contact Sijie Yu at sijie.yu@njit.edu.
    The conversion relies on the Glymur library. Source code and documentation for suncasa.io.ndfits are available at https://github.com/suncasa/suncasa-src.
    Report any code-related issues to the repository maintainers.
    """
    print(hv_comment)
    helioviewer_elem = ET.Element('helioviewer')
    hv_comment_elem = ET.Element('HV_COMMENT')
    hv_comment_elem.text = hv_comment
    helioviewer_elem.append(hv_comment_elem)
    root.append(helioviewer_elem)

    tree._setroot(root)
    return tree
    # from lxml import etree
    # tree = etree.Element("meta")
    # elem = etree.SubElement(tree,"fits")
    # for k,v in header.items():
    #     child = etree.SubElement(elem, k)
    #     if isinstance(v,bool):
    #         v = int(v)
    #     child.text = str(v)
    # return tree


def write_j2000_image(fname, data, header):
    ## todo: write scaled data to jp2. so hv doesn't have to to scale.
    ## todo: add date and time to filename following 2024_03_06__14_18_45_343__EOVSA_1.5GHz.jp2
    ## todo: flare iamges: do similar thing as the CME tag for ccmc data, eOVSa provide movie links to hv. hv will show flare tag
    ## on the solar image and include movie as a external link
    import glymur
    datamax = np.max(data)
    datamin = np.min(data)

    jp2 = glymur.Jp2k(fname + '.tmp.jp2',
                      ((data - datamin) * 256 / (datamax - datamin)).astype(np.uint8), cratios=[20, 10])
    boxes = jp2.box
    header['wavelnth'] = header['crval3']
    header['waveunit'] = header['cunit3']
    header['datamax'] = datamax
    header['datamin'] = datamin
    xmlobj = header_to_xml(header)
    xmlfile = 'image.xml'
    if os.path.exists(xmlfile):
        os.remove(xmlfile)

    xmlobj.write(xmlfile)
    xmlbox = glymur.jp2box.XMLBox(filename='image.xml')
    boxes.insert(3, xmlbox)
    jp2_xml = jp2.wrap(fname, boxes=boxes)

    os.remove(fname + '.tmp.jp2')
    os.remove(xmlfile)
    return True


def wrap(fitsfiles, outfitsfile=None, docompress=False, mask=None, fix_invalid=True, filled_value=0.0, observatory=None,
         imres=None, verbose=False, **kwargs):
    '''
    wrap single frequency fits files into a multiple frequencies fits file
    '''
    from astropy.time import Time
    if len(fitsfiles) <= 1:
        print('There is only one files in the fits file list. wrap is aborted!')
        return ''
    else:
        try:
            fitsfiles = np.array(fitsfiles)
            num_files = len(fitsfiles)
            freqs = np.zeros(num_files)
            for i in range(num_files):
                head = fits.getheader(fitsfiles[i])
                freqs[i] = head['CRVAL3']
                del head
            pos = np.argsort(freqs)
            fitsfiles = fitsfiles[pos]
        except:
            fitsfiles = sorted(fitsfiles)
        nband = len(fitsfiles)
        fits_exist = []
        idx_fits_exist = []
        for sidx, fitsf in enumerate(fitsfiles):
            if os.path.exists(fitsf):
                fits_exist.append(fitsf)
                idx_fits_exist.append(sidx)
        if len(fits_exist) == 0: raise ValueError('None of the input fitsfiles exists!')
        if outfitsfile is None:
            hdu = fits.open(fits_exist[0])
            if observatory is None:
                try:
                    observatory = hdu[-1].header['TELESCOP']
                except:
                    observatory = 'RADIO'
                    print('Failed to acquire telescope information. set as RADIO')
            outfitsfile = Time(hdu[-1].header['DATE-OBS']).strftime(
                '{}.%Y%m%dT%H%M%S.%f.allbd.fits'.format(observatory))
            hdu.close()
        os.system('cp {} {}'.format(fits_exist[0], outfitsfile))
        hdu0 = fits.open(outfitsfile, mode='update')
        header = hdu0[-1].header

        if header['NAXIS'] != 4:
            if verbose:
                print('s1')
            if imres is None:
                cdelts = [325.e8] * nband
                print(
                    'Failed to read bandwidth information. Set as 325 MHz assuming this is EOVSA data. Use the value of keyword CDELT3 with caution.')
            else:
                cdelts = np.squeeze(np.diff(np.array(imres['freq']), axis=1)) * 1e9
            stokesinfo = headerparse(header)
            if stokesinfo['axis'] is None:
                npol = 1
            else:
                npol = int(header['NAXIS{}'.format(stokesinfo['axis'])])
            nbd = nband
            ny = int(header['NAXIS2'])
            nx = int(header['NAXIS1'])

            data = np.zeros((npol, nbd, ny, nx))
            cfreqs = []
            for sidx, fitsf in enumerate(fits_exist):
                hdu = fits.open(fitsf)
                cfreqs.append(hdu[-1].header['RESTFRQ'])
                for pidx in range(npol):
                    if hdu[-1].data.ndim == 2:
                        data[pidx, idx_fits_exist[sidx], :, :] = hdu[-1].data
                    elif hdu[-1].data.ndim == 3:
                        data[pidx, idx_fits_exist[sidx], :, :] = hdu[-1].data[pidx, :, :]
                    else:
                        data[pidx, idx_fits_exist[sidx], :, :] = hdu[-1].data[pidx, 0, :, :]
            cfreqs = np.array(cfreqs)
            cdelts = np.array(cdelts)
            indfreq = np.argsort(cfreqs)
            cfreqs = cfreqs[indfreq]
            cdelts = cdelts[indfreq]
            for pidx in range(npol):
                data[pidx, ...] = data[pidx, indfreq]

            df = np.nanmean(np.diff(cfreqs) / np.diff(idx_fits_exist))  ## in case some of the band is missing
            header['NAXIS'] = 4
            header['NAXIS3'] = nband
            header['CTYPE3'] = 'FREQ'
            header['CRVAL3'] = cfreqs[0]
            header['CDELT3'] = df
            header['CRPIX3'] = 1.0
            header['CUNIT3'] = 'Hz      '
            if stokesinfo['axis'] is None:
                header['NAXIS4'] = 1
                header['CTYPE4'] = 'STOKES'
                header['CRVAL4'] = -5  ## assume XX
                header['CDELT4'] = 1.0
                header['CRPIX4'] = 1.0
                header['CUNIT4'] = '        '
            else:
                for k, v in stokesinfo['headernew'].items():
                    print(k, v)
                    header[k] = v
        else:
            if verbose:
                print('s2')
            npol = int(header['NAXIS4'])
            nbd = nband
            ny = int(header['NAXIS2'])
            nx = int(header['NAXIS1'])

            data = np.zeros((npol, nbd, ny, nx))
            cdelts = []
            cfreqs = []
            cbmaj = []
            cbmin = []
            cbpa = []
            for sidx, fitsf in enumerate(fits_exist):
                hdu = fits.open(fitsf)

                cdelts.append(hdu[-1].header['CDELT3'])
                cfreqs.append(hdu[-1].header['CRVAL3'])
                cbmaj.append(hdu[-1].header['BMAJ'])
                cbmin.append(hdu[-1].header['BMIN'])
                cbpa.append(hdu[-1].header['BPA'])

                for pidx in range(npol):
                    if hdu[-1].data.ndim == 2:
                        data[pidx, idx_fits_exist[sidx], :, :] = hdu[-1].data
                    else:
                        data[pidx, idx_fits_exist[sidx], :, :] = hdu[-1].data[pidx, 0, :, :]

            cfreqs = np.array(cfreqs)
            cdelts = np.array(cdelts)
            cbmaj = np.array(cbmaj)
            cbmin = np.array(cbmin)
            cbpa = np.array(cbpa)

            df = np.nanmean(np.diff(cfreqs) / np.diff(idx_fits_exist))  ## in case some of the band is missing
            header['cdelt3'] = df
            header['NAXIS3'] = nband
            header['NAXIS'] = 4
            header['CRVAL3'] = header['CRVAL3'] - df * idx_fits_exist[0]

        for dim1 in range(1, header['NAXIS'] + 1):
            for dim2 in range(1, header['NAXIS'] + 1):
                k = 'PC{:02d}_{:d}'.format(dim1, dim2)
                if dim1 == dim2:
                    header[k] = 1.0
                else:
                    header[k] = 0.0

        if os.path.exists(outfitsfile):
            os.system('rm -rf {}'.format(outfitsfile))

        col1 = fits.Column(name='cfreqs', format='E', array=cfreqs)
        col2 = fits.Column(name='cdelts', format='E', array=cdelts)
        col3 = fits.Column(name='bmaj', format='E', array=cbmaj)
        col4 = fits.Column(name='bmin', format='E', array=cbmin)
        col5 = fits.Column(name='bpa', format='E', array=cbpa)
        tbhdu = fits.BinTableHDU.from_columns([col1, col2, col3, col4, col5])

        if docompress:
            if fix_invalid:
                data[np.isnan(data)] = filled_value
            if kwargs is {}:
                kwargs.update({'compression_type': 'RICE_1', 'quantize_level': 4.0})
            if isinstance(outfitsfile, str):
                outfitsfile = os.path.expanduser(outfitsfile)

            header, data = headersqueeze(header, data)
            if data.ndim == 4:
                print('only 1D, 2D, or 3D images are currently supported for compression. Aborting compression...')
            else:
                hdunew = fits.CompImageHDU(data=data, header=header, **kwargs)

                if mask is None:
                    hdulnew = fits.HDUList([fits.PrimaryHDU(), hdunew, tbhdu])
                else:
                    hdumask = fits.CompImageHDU(data=mask.astype(np.uint8), **kwargs)
                    hdulnew = fits.HDUList([fits.PrimaryHDU(), hdunew, tbhdu, hdumask])
                hdulnew.writeto(outfitsfile, output_verify='fix')
                return outfitsfile

        hdulnew = fits.HDUList([fits.PrimaryHDU(data=data, header=header), tbhdu])
        hdulnew.writeto(outfitsfile)
        print('wrapped fits written as ' + outfitsfile)
        return outfitsfile


def update(fitsfile, new_data=None, new_columns=None, new_header_entries=None):
    """
    Updates a FITS file by optionally replacing its primary or compressed image data, adding new columns to the
    first binary table (BinTableHDU), and/or updating header entries in the first image HDU (PrimaryHDU for
    uncompressed or CompImageHDU for compressed FITS files).

    Parameters:
    - fitsfile (str): Path to the FITS file to be updated.
    - new_data (np.ndarray, optional): New data array to replace the existing data in the first image HDU.
      Defaults to None, which means the data will not be updated.
    - new_columns (list of astropy.io.fits.Column, optional): New columns to be added to the first BinTableHDU.
      Defaults to None, which means no columns will be added.
    - new_header_entries (dict, optional): Header entries to update or add in the first image HDU. Each key-value
      pair represents a header keyword and its new value. Defaults to None, which means no header updates will be made.

    Returns:
    - bool: True if any of the specified updates were successfully applied, False otherwise.

    The function determines whether the FITS file is compressed to properly handle the image HDU type. It attempts
    to update the image HDU's data, the BinTableHDU's columns, and the image HDU's header based on the provided
    arguments. If all input parameters are None, indicating no updates are specified, the function will print a message
    and return False.
    """
    if new_data is None and new_columns is None and new_header_entries is None:
        print("No updates to perform: all new_data new_columns and new_header_entries are None.")
        return False

    if is_compressed_fits(fitsfile):
        imagehdutype = fits.hdu.CompImageHDU
    else:
        imagehdutype = fits.hdu.PrimaryHDU

    try:
        with fits.open(fitsfile, mode='update') as hdul:
            if new_header_entries is None:
                header_updated = True
                ## skip updating header
            else:
                header_updated = False
                for hdu in hdul:
                    if isinstance(hdu, imagehdutype):
                        for key, value in new_header_entries.items():
                            if key == "HISTORY":
                                hdu.header.add_history(value)
                            else:
                                hdu.header[key] = value
                        header_updated = True
                        break  # Updating only the first relevant image HDU

            if new_data is None:
                data_updated = True
            else:
                data_updated = False
                for hdu in hdul:
                    if isinstance(hdu, imagehdutype):
                        hdu.data = new_data
                        data_updated = True
                        break
            if new_columns is None:
                bintable_updated = True
                ## skip updating bintable
            else:
                # Add new columns to the first BinTableHDU
                bintable_updated = False
                for idx, hdu in enumerate(hdul):
                    if isinstance(hdu, imagehdutype):
                        continue  # Skip compressed image HDUs
                    if isinstance(hdu, fits.hdu.BinTableHDU):
                        combined_columns = fits.ColDefs(hdu.columns) + fits.ColDefs(new_columns)
                        new_tbhdu = fits.BinTableHDU.from_columns(combined_columns)
                        hdul[idx] = new_tbhdu
                        bintable_updated = True
                        break  # Updating only the first BinTableHDU
            if not header_updated:
                print("Header update failed.")
            if not bintable_updated:
                print("BinTable update failed.")
            if not data_updated:
                print("Data update failed.")
            if not header_updated and not bintable_updated and not data_updated:
                return False
            return True
    except Exception as e:
        print(f"An error occurred: {e}")
        return False



from ...casa_compat import check_dependencies
check_dependencies()

import sys
import os
import numpy as np
import numpy.ma as ma
import scipy.constants as constants
import time
from astropy.time import Time
from eovsapy import util
from eovsapy import dump_tsys as dtsys
import aipy
from suncasa.eovsa import impteovsa as ipe

py3 = sys.version_info.major >= 3

from ...casa_compat import import_casatools, import_casatasks

tasks = import_casatasks('split', 'casalog')
split = tasks.get('split')
casalog = tasks.get('casalog')

tools = import_casatools(['tbtool', 'mstool', 'qatool', 'iatool'])
tbtool = tools['tbtool']
mstool = tools['mstool']
qatool = tools['qatool']
iatool = tools['iatool']
tb = tbtool()
ms = mstool()
qa = qatool()
ia = iatool()

c_external = False


# idbdir = os.getenv('EOVSAIDB')
#
# if not idbdir:
#     print('Environmental variable for EOVSA idb path not defined')
#     print('Use default path on pipeline')
#     idbdir = '/data1/eovsa/fits/IDB/'


def udb_corr_external(filelist, udbcorr_path, use_exist_udbcorr=False):
    import pickle
    udbcorr_script = os.path.join(udbcorr_path, 'udbcorr_ext.py')
    if os.path.exists(udbcorr_script):
        os.system('rm -rf {}'.format(udbcorr_script))
    udbcorr_file = os.path.join(udbcorr_path, 'udbcorr_tmp.pickle')

    if use_exist_udbcorr and os.path.exists(udbcorr_file):
        with open(udbcorr_file, 'rb') as sf:
            filelist = pickle.load(sf)
    else:
        if os.path.exists(udbcorr_file):
            os.system('rm -rf {}'.format(udbcorr_file))
        with open(udbcorr_file, 'wb') as sf:
            pickle.dump(filelist, sf)
        fi = open(udbcorr_script, 'wb')
        fi.write(b'import pickle \n')
        fi.write(b'import pipeline_cal as pc \n')
        fi.write(b'import sys \n')
        fi.write(b'syspath = sys.path \n')
        fi.write(b"sys.path = [l for l in syspath if 'casa' not in l] \n")
        line = "with open('{}', 'rb') as sf: \n".format(udbcorr_file)
        fi.write(line.encode())
        fi.write(b'    filelist = pickle.load(sf) \n')

        fi.write(b'filelist_tmp = [] \n')
        fi.write(b'for ll in filelist: \n')
        fi.write(b"    try: \n")
        line = "        filelist_tmp.append(pc.udb_corr(ll, outpath='{}/', calibrate=True, desat=True)) \n".format(
            udbcorr_path)
        fi.write(line.encode())
        fi.write(b"    except: \n")
        fi.write(b"        pass \n")
        fi.write(b'filelist = filelist_tmp \n')
        line = "with open('{}', 'wb') as sf: \n".format(udbcorr_file)
        fi.write(line.encode())
        fi.write(b'    pickle.dump(filelist,sf) \n')
        fi.close()

        udbcorr_shellscript = os.path.join(udbcorr_path, 'udbcorr_ext.csh')
        if os.path.exists(udbcorr_shellscript):
            os.system('rm -rf {}'.format(udbcorr_shellscript))
        fi = open(udbcorr_shellscript, 'wb')
        fi.write(b'#! /bin/tcsh -f \n')
        fi.write(b' \n')
        # fi.write('setenv PYTHONPATH "/home/user/test_svn/python:/common/python/current:/common/python" \n')
        fi.write(b'source /home/user/.cshrc \n')
        line = '/common/anaconda2/bin/python {} \n'.format(udbcorr_script)
        fi.write(line.encode())
        fi.close()

        os.system('/bin/tcsh {}'.format(udbcorr_shellscript))

        with open(udbcorr_file, 'rb') as sf:
            filelist = pickle.load(sf)

    if filelist == []:
        raise ValueError('udb_corr failed to return any results. Please check your calibration.')
    return filelist


def trange2filelist(trange=[], verbose=False):
    '''This finds all solar IDB files within a timerange;
       Required inputs:
       trange - can be 1) a single string or Time() object in UTC: use the entire day, e.g., '2017-08-01' or Time('2017-08-01')
                          if just a date, find all scans withing the same date in local time.
                          if a complete time stamp, find the local date first (which may be different from that provided,
                            and return all scans within that day
                       2) a range of Time(), e.g., Time(['2017-08-01 00:00','2017-08-01 23:00'])
                       3) None -- use current date Time.now()
    '''

    if trange:
        if type(trange) == list or type(trange) == str:
            try:
                trange = Time(trange)
            except:
                print('trange format not recognised. Abort....')
                return None
    else:
        print('Please give a time range. Abort....')
        return None
    # if type(trange) == Time:
    try:
        # if single Time object, the following line would report an error
        nt = len(trange)
        if len(trange) > 1:
            # more than one value
            trange = Time([trange[0], trange[-1]])
        else:
            # single value in a list
            trange = Time(np.array([-1.0, 1.0]) * 5 / 24. / 60. + trange[0].mjd, format='mjd')
    except:
        trange = Time(np.array([-1.0, 1.0]) * 5 / 24. / 60. + trange.mjd, format='mjd')

    t1 = trange[0].datetime
    t2 = trange[1].datetime
    daydelta = (t2.date() - t1.date()).days
    if t1.date() != t2.date():
        # End day is different than start day, so read and concatenate two fdb files
        info = dtsys.rd_fdb(trange[0])
        for ll in range(daydelta):
            info2 = dtsys.rd_fdb(Time(trange[0].mjd + ll + 1, format='mjd'))
            if info2:
                for key in info.keys():
                    info.update({key: np.append(info[key], info2[key])})
    else:
        # Both start and end times are on the same day
        info = dtsys.rd_fdb(trange[0])

    # remove empty items
    for k, v in info.items():
        info[k] = info[k][~(info[k] == '')]

    sidx = np.where(
        np.logical_and(info['SOURCEID'] == 'Sun', info['PROJECTID'] == 'NormalObserving') & np.logical_and(
            info['ST_TS'].astype(float) >= trange[0].lv,
            info['ST_TS'].astype(float) <= trange[
                1].lv))
    filelist = info['FILE'][sidx]
    if verbose:
        print(
            '{} file found in the time range from {} to {}: '.format(len(filelist), t1.strftime('%Y-%m-%d %H:%M:%S UT'),
                                                                     t2.strftime('%Y-%m-%d %H:%M:%S UT')))
    idbdir = util.get_idbdir(Time(t1))
    inpath = '{}/{}/'.format(idbdir, trange[0].datetime.strftime("%Y%m%d"))
    filelist = [inpath + ll for ll in filelist]
    return filelist


def importeovsa_iter(filelist, timebin, width, visprefix, nocreatms, modelms, doscaling, keep_nsclms, fileidx):
    '''

    '''

    filename = filelist[fileidx]
    uv = aipy.miriad.UV(filename)
    # try:
    msname0 = list(filename.split('/')[-1])
    msname = visprefix + ''.join(msname0) + '.ms'
    # try:
    # uv.select('antennae', 0, 1, include=True)
    # uv.select('polarization', -5, -5, include=True)

    if 'antlist' in uv.vartable:
        ants = uv['antlist'].replace('\x00', '')
        antlist = list(map(int, ants.split()))
    else:
        antlist = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16]

    good_idx = np.where(uv['sfreq'] > 0)[0]

    nf = len(good_idx)
    npol = uv['npol']
    nants = uv['nants']
    source_id = uv['source'].replace('\x00', '')
    sfreq = uv['sfreq'][good_idx]
    sdf = uv['sdf'][good_idx]
    ra, dec = uv['ra'], uv['dec']
    nbl = int(nants * (nants - 1) / 2)
    bl2ord = ipe.bl_list2(nants)
    npairs = int(nbl + nants)

    timesall = []
    uv.rewind()
    for preamble, data in uv.all():
        uvw, t, (i, j) = preamble
        timesall.append(t)
    timesjd = np.unique(timesall)
    # except:
    #     pass

    uv.select('clear', -1, -1, include=True)
    times = ipe.jd2mjds(np.asarray(timesjd))
    inttime = np.median((times - np.roll(times, 1))[1:]) / 60  ## time in minutes
    inttimed = inttime / (24 * 60)  ## time in days

    time_steps = np.round((times[-1] - times[0]) / inttime / 60).astype(int) + 1
    # time_steps = len(timesall) / (npairs * npol)
    if len(times) != time_steps:
        ### This is to solve the timestamp glitch in idb files.
        ### The timestamps are supposed to be evenly spaced
        ### However, some idb files may miss a few timestamps in the evenly-spaced time grid.
        ### The step will map the the data to the evenly-spaced time grid.
        timesnew = np.linspace(times[0], times[-1], time_steps)
        timesnew[np.hstack([[0], np.cumsum(np.round(np.diff(times) / 60 / inttime))]).astype(int)] = times
        times = timesnew
    durtim = int(np.round((times[-1] - times[0]) / 60 + inttime))  ## time in minutes
    time0 = time.time()

    flag = np.ones((npol, nf, time_steps, npairs), dtype=bool)
    out = np.zeros((npol, nf, time_steps, npairs), dtype=np.complex64)  # Cross-correlations
    uvwarray = np.zeros((3, time_steps, npairs), dtype=float)
    chan_band = ipe.get_band(sfreq=sfreq, sdf=sdf, date=Time(uv['time'], format='jd'))
    nband = len(chan_band)

    uv.rewind()
    l = -1
    for preamble, data in uv.all():
        uvw, t, (i0, j0) = preamble
        i = antlist.index(i0 + 1)
        j = antlist.index(j0 + 1)
        if i > j:
            # Reverse order of indices
            j = antlist.index(i0 + 1)
            i = antlist.index(j0 + 1)
        # Assumes uv['pol'] is one of -5, -6, -7, -8
        k = -5 - uv['pol']
        l += 1
        mask0 = data.mask
        data = ma.masked_array(ma.masked_invalid(data), fill_value=0.0)
        try:
            tidx = np.where(np.abs(timesjd - t) < inttimed)[0][0]
        except:
            tidx = l / (npairs * npol)
        out[k, :, tidx, bl2ord[i0, j0]] = data.data
        flag[k, :, tidx, bl2ord[i0, j0]] = np.logical_or(data.mask, mask0)
        # if i != j:
        if k == 3:
            uvwarray[:, tidx, bl2ord[i0, j0]] = -uvw * constants.speed_of_light / 1e9

    nrows = time_steps * npairs
    if doscaling:
        out2 = out.copy()
        for i0 in antlist:
            for j0 in antlist:
                if i0 < j0:
                    i, j = i0 - 1, j0 - 1
                    out2[:, :, :, bl2ord[i, j]] = out[:, :, :, bl2ord[i, j]] / np.sqrt(
                        np.abs(out[:, :, :, bl2ord[i, i]]) * np.abs(out[:, :, :, bl2ord[j, j]]))
        out2 = out2.reshape(npol, nf, nrows)
        out2[np.isnan(out2)] = 0
        out2[np.isinf(out2)] = 0
    # out2 = ma.masked_array(ma.masked_invalid(out2), fill_value=0.0)
    out = out.reshape(npol, nf, nrows) * 1e4
    flag = flag.reshape(npol, nf, nrows)
    uvwarray = uvwarray.reshape(3, nrows)
    uvwarray = np.tile(uvwarray, (1, nband))
    sigma = np.ones((4, nrows), dtype=float) + 1
    sigma = np.tile(sigma, (1, nband))

    casalog.post('IDB File {0} is readed in --- {1:10.2f} seconds ---'.format(filename, (time.time() - time0)))

    if not nocreatms:
        modelms = ipe.creatms(filename, visprefix)
        os.system('mv {} {}'.format(modelms, msname))
    else:
        casalog.post('----------------------------------------')
        casalog.post('copying standard MS to {0}'.format(msname, (time.time() - time0)))
        casalog.post('----------------------------------------')
        os.system("rm -fr {}".format(msname))
        os.system("cp -r {} {}".format(modelms, msname))
        casalog.post('Standard MS is copied to {0} in --- {1:10.2f} seconds ---'.format(msname, (time.time() - time0)))

    tb.open(msname, nomodify=False)
    casalog.post('----------------------------------------')
    casalog.post("Updating the main table of {}".format(msname))
    casalog.post('----------------------------------------')
    for l, cband in enumerate(chan_band):
        time1 = time.time()
        # nchannels = len(cband['cidx'])
        for row in range(nrows):
            if not doscaling or keep_nsclms:
                datacell = out[:, cband['cidx'][0]:cband['cidx'][-1] + 1, row]
                rownr = int(row + l * nrows)
                tb.putcell('DATA', rownr, datacell)
            tb.putcell('FLAG', rownr, flag[:, cband['cidx'][0]:cband['cidx'][-1] + 1, row])
        casalog.post('---spw {0:02d} is updated in --- {1:10.2f} seconds ---'.format((l + 1), time.time() - time1))
    tb.putcol('UVW', uvwarray)
    tb.putcol('SIGMA', sigma)
    tb.putcol('WEIGHT', 1.0 / sigma ** 2)
    timearr = times
    timearr = timearr.reshape(1, time_steps, 1)
    timearr = np.tile(timearr, (nband, 1, npairs))
    timearr = timearr.reshape(nband * npairs * time_steps)
    tb.putcol('TIME', timearr)
    tb.putcol('TIME_CENTROID', timearr)
    scan_id = tb.getcol('SCAN_NUMBER')
    scan_id *= 0
    tb.putcol('SCAN_NUMBER', scan_id)
    colnames = tb.colnames()
    cols2rm = ["MODEL_DATA", "CORRECTED_DATA"]
    for l in range(len(cols2rm)):
        if cols2rm[l] in colnames:
            tb.removecols(cols2rm[l])
    tb.close()

    casalog.post('----------------------------------------')
    casalog.post("Updating the OBSERVATION table of {}".format(msname))
    casalog.post('----------------------------------------')
    tb.open(msname + '/OBSERVATION', nomodify=False)
    tb.putcol('TIME_RANGE', np.asarray([times[0] - 0.5 * inttime, times[-1] + 0.5 * inttime]).reshape(2, 1))
    tb.putcol('OBSERVER', ['EOVSA team'])
    tb.close()

    casalog.post('----------------------------------------')
    casalog.post("Updating the POINTING table of {}".format(msname))
    casalog.post('----------------------------------------')
    tb.open(msname + '/POINTING', nomodify=False)
    timearr = times.reshape(1, time_steps, 1)
    timearr = np.tile(timearr, (nband, 1, nants))
    timearr = timearr.reshape(nband * time_steps * nants)
    tb.putcol('TIME', timearr)
    tb.putcol('TIME_ORIGIN', timearr)  # - 0.5 * delta_time)
    direction = tb.getcol('DIRECTION')
    direction[0, 0, :] = ra
    direction[1, 0, :] = dec
    tb.putcol('DIRECTION', direction)
    target = tb.getcol('TARGET')
    target[0, 0, :] = ra
    target[1, 0, :] = dec
    tb.putcol('TARGET', target)
    tb.close()

    casalog.post('----------------------------------------')
    casalog.post("Updating the SOURCE table of {}".format(msname))
    casalog.post('----------------------------------------')
    tb.open(msname + '/SOURCE', nomodify=False)
    radec = tb.getcol('DIRECTION')
    radec[0], radec[1] = ra, dec
    tb.putcol('DIRECTION', radec)
    name = np.array([source_id], dtype='|S{0}'.format(len(source_id) + 1))
    tb.putcol('NAME', name)
    tb.close()

    casalog.post('----------------------------------------')
    casalog.post("Updating the DATA_DESCRIPTION table of {}".format(msname))
    casalog.post('----------------------------------------')
    tb.open(msname + '/DATA_DESCRIPTION/', nomodify=False)
    pol_id = tb.getcol('POLARIZATION_ID')
    pol_id *= 0
    tb.putcol('POLARIZATION_ID', pol_id)
    # spw_id = tb.getcol('SPECTRAL_WINDOW_ID')
    # spw_id *= 0
    # tb.putcol('SPECTRAL_WINDOW_ID', spw_id)
    tb.close()

    # casalog.post('----------------------------------------')
    # casalog.post("Updating the POLARIZATION table of {}".format(msname))
    # casalog.post('----------------------------------------')
    # tb.open(msname + '/POLARIZATION/', nomodify=False)
    # tb.removerows(rownrs=np.arange(1, nband, dtype=int))
    # tb.close()

    casalog.post('----------------------------------------')
    casalog.post("Updating the FIELD table of {}".format(msname))
    casalog.post('----------------------------------------')
    tb.open(msname + '/FIELD/', nomodify=False)
    delay_dir = tb.getcol('DELAY_DIR')
    delay_dir[0], delay_dir[1] = ra, dec
    tb.putcol('DELAY_DIR', delay_dir)
    phase_dir = tb.getcol('PHASE_DIR')
    phase_dir[0], phase_dir[1] = ra, dec
    tb.putcol('PHASE_DIR', phase_dir)
    reference_dir = tb.getcol('REFERENCE_DIR')
    reference_dir[0], reference_dir[1] = ra, dec
    tb.putcol('REFERENCE_DIR', reference_dir)
    name = np.array([source_id], dtype='|S{0}'.format(len(source_id) + 1))
    tb.putcol('NAME', name)
    tb.close()

    # FIELD: DELAY_DIR, PHASE_DIR, REFERENCE_DIR, NAME

    # del out, flag, uvwarray, uv, timearr, sigma
    # gc.collect()  #
    if doscaling:
        if keep_nsclms:
            msname_scl = visprefix + ''.join(msname0) + '_scl.ms'
            os.system('cp -r {} {}'.format(msname, msname_scl))
        else:
            msname_scl = msname
        tb.open(msname_scl, nomodify=False)
        casalog.post('----------------------------------------')
        casalog.post("Updating the main table of {}".format(msname_scl))
        casalog.post('----------------------------------------')
        for l, cband in enumerate(chan_band):
            time1 = time.time()
            for row in range(nrows):
                rownr = int(row + l * nrows)
                tb.putcell('DATA', rownr, out2[:, cband['cidx'][0]:cband['cidx'][-1] + 1, row])
            casalog.post('---spw {0:02d} is updated in --- {1:10.2f} seconds ---'.format((l + 1), time.time() - time1))
        tb.close()

    if not (timebin == '0s' and width == 1):
        msfile = msname + '.split'
        if doscaling:
            split(vis=msname_scl, outputvis=msname_scl + '.split', datacolumn='data', timebin=timebin, width=width,
                  keepflags=False)
            os.system('rm -rf {}'.format(msname_scl))
            msfile_scl = msname_scl + '.split'
        if not (doscaling and not keep_nsclms):
            split(vis=msname, outputvis=msname + '.split', datacolumn='data', timebin=timebin, width=width,
                  keepflags=False)
            os.system('rm -rf {}'.format(msname))
    else:
        msfile = msname
        if doscaling:
            msfile_scl = msname_scl
    casalog.post("finished in --- {:.1f} seconds ---".format(time.time() - time0))
    if doscaling:
        return [True, msfile, msfile_scl, durtim]
    else:
        return [True, msfile, durtim]


def importeovsa(idbfiles=None, ncpu=None, timebin=None, width=None, visprefix=None, udb_corr=True, nocreatms=None,
                doconcat=None, modelms=None,
                doscaling=False, keep_nsclms=False, use_exist_udbcorr=False):
    casalog.origin('importeovsa')

    if type(idbfiles) == Time:
        filelist = trange2filelist(idbfiles)
    else:
        # If input type is not Time, assume that it is the list of files to read
        filelist = idbfiles

    if type(filelist) == str:
        filelist = [filelist]

    filelist_tmp = []
    for ll in filelist:
        if not os.path.exists(ll):
            casalog.post("Warning: {} not exist.".format(ll))
        else:
            filelist_tmp.append(ll)

    filelist = filelist_tmp
    if not filelist:
        casalog.post("No file in idbfiles list exists. Abort.")
        return False

    for idx, ll in enumerate(filelist):
        if ll[-1] == '/':
            filelist[idx] = ll[:-1]

    if not visprefix:
        visprefix = './'
    else:
        if os.path.exists(visprefix):
            pass
        else:
            casalog.post("The output path {} does not exist. Abort.".format(visprefix))
            return False
    if not timebin:
        timebin = '0s'
    if not width:
        width = 1

    if udb_corr:
        from eovsapy.pipeline_cal import udb_corr
        udbcorr_path = visprefix + '/tmp_UDBcorr/'
        if not os.path.exists(udbcorr_path):
            os.makedirs(udbcorr_path)
        filelist_tmp = []
        for ll in filelist:
            try:
                filelist_tmp.append(udb_corr(ll, outpath='{}/'.format(udbcorr_path), calibrate=True, desat=True))
            except:
                pass

        if filelist_tmp == []:
            raise ValueError('udb_corr failed to return any results. Please check your calibration.')
        else:
            filelist = filelist_tmp
        # filelist = udb_corr_external(filelist, udbcorr_path, use_exist_udbcorr)

    if not modelms:
        if nocreatms:
            filename = filelist[0]
            modelms = ipe.creatms(filename, visprefix)
    else:
        if not os.path.exists(modelms):
            if nocreatms:
                filename = filelist[0]
                modelms = ipe.creatms(filename, visprefix)

    iterable = range(len(filelist))

    t0 = time.time()
    casalog.post('Perform importeovsa in parallel with {} CPUs...'.format(ncpu))

    if ncpu == 1:
        res = []
        for fidx, ll in enumerate(filelist):
            res.append(
                importeovsa_iter(filelist, timebin, width, visprefix, nocreatms, modelms, doscaling, keep_nsclms, fidx))
    if ncpu > 1:
        import multiprocessing as mprocs
        from functools import partial
        imppart = partial(importeovsa_iter, filelist, timebin, width, visprefix, nocreatms, modelms, doscaling,
                          keep_nsclms)
        pool = mprocs.Pool(ncpu)
        res = pool.map(imppart, iterable)
        pool.close()
        pool.join()

    # print res
    t1 = time.time()
    timelapse = t1 - t0
    print('It took %f secs to complete' % timelapse)

    # results = pd.DataFrame({'succeeded': [], 'msfile': [], 'durtim': []})
    # for r in res:
    #     results = results.append(pd.DataFrame({'succeeded': [r[0]], 'msfile': [r[1]], 'durtim': [r[2]]}))
    # try:
    succeeded = []
    msfile = []
    durtim = []
    if doscaling:
        msfile_scl = []
        for r in res:
            succeeded.append(r[0])
            msfile.append(r[1])
            msfile_scl.append(r[2])
            durtim.append(r[3])
        results = {'succeeded': succeeded, 'msfile': msfile, 'msfile_scl': msfile_scl, 'durtim': durtim}
    else:
        for r in res:
            succeeded.append(r[0])
            msfile.append(r[1])
            durtim.append(r[2])
        results = {'succeeded': succeeded, 'msfile': msfile, 'durtim': durtim}
    # except:
    #     print 'errors occurred when creating the output summary.'

    if udb_corr:
        os.system('rm -rf {}'.format(udbcorr_path))

    if doconcat:
        from suncasa.suncasatasks.private import task_concateovsa as ce
        msname = os.path.basename(filelist[0])
        durtim = int(np.array(results['durtim']).sum())
        if doscaling:
            msfiles = list(np.array(results['msfile_scl'])[np.where(np.array(results['succeeded']) == True)])
            if keep_nsclms:
                concatvis = visprefix + msname + '-{:d}m{}.ms'.format(durtim, '_scl')
            else:
                concatvis = visprefix + msname + '-{:d}m{}.ms'.format(durtim, '')
        else:
            msfiles = list(np.array(results['msfile'])[np.where(np.array(results['succeeded']) == True)])
            concatvis = visprefix + msname + '-{:d}m{}.ms'.format(durtim, '')
        ce.concateovsa(msfiles, concatvis, datacolumn='data', keep_orig_ms=True, cols2rm="model,corrected")
        return concatvis
    else:
        msfiles = list(np.array(results['msfile'])[np.where(np.array(results['succeeded']) == True)])
        return [str(m) for m in msfiles]



import multiprocessing
from suncasa.utils import qlookplot as ql
import platform

msfile = 'IDB20210507_1840-1920XXYY.cal.10s.slfcaled.ms'
specfile = msfile + '.dspec.npz'
## set the time interval
timerange = '2021/05/07/19:01:30~2021/05/07/19:03:00'
## Bin width for time averaging
twidth = 1
## frequency range
spw=['0~1','5','10','15','25','30','40~45']
## image center for clean in solar X-Y in arcsec
xycen = [-900, 280]
## pixel scale
cell = ['2.0arcsec']
## number of pixels in X and Y. If only one value is provided, NX = NY
imsize = [128]
## field of view of the zoomed-in panels in unit of arcsec
fov = [300, 300]
## select stokes XX
stokes = 'XX'
## set True if make a movie
mkmovie = True
## set True if generate compressed fits
docompress = True
## numbers of CPU threads for computing
if platform.system() == 'Darwin':
    ncpu = multiprocessing.cpu_count()
else:
    ncpu = 1  # multiprocessing.cpu_count()
## set False to plot EOVSA images as open contours, set True to plot as filled contours
opencontour = False
clevels = [0.5, 1.0]

#### ---- Control knobs for AIA plotting ---- ####
## set True if plot AIA images as the background
plotaia = True
## AIA passband. The options are [171,131,304,335,211,193,94,1600,1700]
aiawave = 1600

movieformat = 'mp4'

ql.qlookplot(vis=msfile, specfile=specfile, timerange=timerange, spw=spw,
             ncpu=ncpu, xycen=xycen, imsize=imsize, fov=fov, cell=cell,
             restoringbeam=['60arcsec'],robust=0.5,
             opencontour=opencontour, clevels=clevels,
             plotaia=plotaia, aiawave=aiawave,
             mkmovie=mkmovie, twidth=twidth, docompress=docompress, stokes=stokes,
             movieformat='mp4', overwrite=False)



## This code is obsolete. It is replaced by task_ptclean6.py


import os
import numpy as np
import shutil
from functools import partial
from time import time
import glob
import sys
from ...utils import helioimage2fits as hf
from ...casa_compat import import_casatools, import_casatasks

tasks = import_casatasks('split', 'tclean', 'casalog')
split = tasks.get('split')
tclean = tasks.get('tclean')
casalog = tasks.get('casalog')

tools = import_casatools(['tbtool', 'mstool', 'qatool'])
tbtool = tools['tbtool']
mstool = tools['mstool']
qatool = tools['qatool']
tb = tbtool()
ms = mstool()
qa = qatool()

c_external = False


def clean_iter(tim, vis, imageprefix, imagesuffix,
               twidth, doreg, docompress, usephacenter, reftime, ephem, msinfo, toTb, sclfactor, overwrite,
               selectdata, field, spw,
               uvrange, antenna, scan, observation, intent, datacolumn, imsize, cell, phasecenter, stokes,
               projection, startmodel, specmode, reffreq, nchan, start, width, outframe, veltype, restfreq,
               interpolation, gridder, facets, chanchunks, wprojplanes, vptable, usepointing, mosweight, aterm, psterm,
               wbawp, conjbeams, cfcache, computepastep, rotatepastep, pblimit, normtype, deconvolver, scales, nterms,
               smallscalebias, restoration, restoringbeam, pbcor, outlierfile, weighting, robust, npixels, uvtaper,
               niter, gain, threshold, nsigma, cycleniter, cyclefactor, minpsffraction, maxpsffraction, interactive,
               usemask, mask, pbmask, sidelobethreshold, noisethreshold, lownoisethreshold, negativethreshold,
               smoothfactor, minbeamfrac, cutthreshold, growiterations, dogrowprune, minpercentchange, verbose, restart,
               savemodel, calcres, calcpsf, parallel, subregion, tmpdir, btidx):
    bt = btidx  # 0
    if bt + twidth < len(tim) - 1:
        et = btidx + twidth - 1
    else:
        et = len(tim) - 1

    if bt == 0:
        bt_d = tim[bt] - ((tim[bt + 1] - tim[bt]) / 2)
    else:
        bt_d = tim[bt] - ((tim[bt] - tim[bt - 1]) / 2)
    if et == (len(tim) - 1) or et == -1:
        et_d = tim[et] + ((tim[et] - tim[et - 1]) / 2)
    else:
        et_d = tim[et] + ((tim[et + 1] - tim[et]) / 2)

    timerange = qa.time(qa.quantity(bt_d, 's'), prec=9, form='ymd')[0] + '~' + \
                qa.time(qa.quantity(et_d, 's'), prec=9, form='ymd')[0]
    btstr = qa.time(qa.quantity(bt_d, 's'), prec=9, form='fits')[0]
    etstr = qa.time(qa.quantity(et_d, 's'), prec=9, form='fits')[0]
    print('cleaning timerange: ' + timerange)

    image0 = btstr.replace(':', '').replace('-', '')
    imname = imageprefix + image0 + imagesuffix

    if overwrite or (len(glob.glob(imname + '*')) == 0):
        os.system('rm -rf {}*'.format(imname))
        try:
            tclean(vis=vis, selectdata=selectdata, field=field, spw=spw, timerange=timerange, uvrange=uvrange,
                   antenna=antenna, scan=scan, observation=observation, intent=intent, datacolumn=datacolumn,
                   imagename=imname, imsize=imsize, cell=cell, phasecenter=phasecenter, stokes=stokes,
                   projection=projection, startmodel=startmodel, specmode=specmode, reffreq=reffreq, nchan=nchan,
                   start=start, width=width, outframe=outframe, veltype=veltype, restfreq=restfreq,
                   interpolation=interpolation, gridder=gridder, facets=facets, chanchunks=chanchunks,
                   wprojplanes=wprojplanes, vptable=vptable, usepointing=usepointing, mosweight=mosweight, aterm=aterm,
                   psterm=psterm, wbawp=wbawp, conjbeams=conjbeams, cfcache=cfcache, computepastep=computepastep,
                   rotatepastep=rotatepastep, pblimit=pblimit, normtype=normtype, deconvolver=deconvolver,
                   scales=scales, nterms=nterms, smallscalebias=smallscalebias, restoration=restoration,
                   restoringbeam=restoringbeam, pbcor=pbcor, outlierfile=outlierfile, weighting=weighting,
                   robust=robust, npixels=npixels, uvtaper=uvtaper, niter=niter, gain=gain, threshold=threshold,
                   nsigma=nsigma, cycleniter=cycleniter, cyclefactor=cyclefactor, minpsffraction=minpsffraction,
                   maxpsffraction=maxpsffraction, interactive=interactive, usemask=usemask, mask=mask, pbmask=pbmask,
                   sidelobethreshold=sidelobethreshold, noisethreshold=noisethreshold,
                   lownoisethreshold=lownoisethreshold, negativethreshold=negativethreshold, smoothfactor=smoothfactor,
                   minbeamfrac=minbeamfrac, cutthreshold=cutthreshold, growiterations=growiterations,
                   dogrowprune=dogrowprune, minpercentchange=minpercentchange, verbose=verbose, restart=restart,
                   savemodel=savemodel, calcres=calcres, calcpsf=calcpsf, parallel=parallel)
            # print('checkpoint 3')
            if pbcor:
                clnjunks = ['.flux', '.mask', '.model', '.psf', '.residual', '.pb', '.sumwt', '.image']
            else:
                clnjunks = ['.flux', '.mask', '.model', '.psf', '.residual', '.pb', '.sumwt', '.image.pbcor']
            for clnjunk in clnjunks:
                if os.path.exists(imname + clnjunk):
                    shutil.rmtree(imname + clnjunk)
            if pbcor:
                os.system('mv {} {}'.format(imname + '.image.pbcor', imname + '.image'))
        except:
            print('error in cleaning image: ' + btstr)
            return [False, btstr, etstr, '']
    else:
        print(imname + ' exists. Clean task aborted.')

    if doreg:
        # ephem.keys()
        # msinfo.keys()
        if os.path.isfile(imname + '.fits'):
            return [True, btstr, etstr, imname + '.fits']
        else:
            try:
                # check if ephemfile and msinfofile exist
                if not ephem:
                    print("ephemeris info does not exist, querying from JPL Horizons on the fly")
                    ephem = hf.read_horizons(vis=vis)
                if not msinfo:
                    print("ms info not provided, generating one on the fly")
                    msinfo = hf.read_msinfo(vis)
                hf.imreg(vis=vis, ephem=ephem, msinfo=msinfo, timerange=timerange, reftime=reftime,
                         imagefile=imname + '.image', fitsfile=imname + '.fits', overwrite=True,
                         toTb=toTb, sclfactor=sclfactor, usephacenter=usephacenter, subregion=subregion,
                         docompress=docompress)
                if os.path.exists(imname + '.fits'):
                    shutil.rmtree(imname + '.image')
                    return [True, btstr, etstr, imname + '.fits']
                else:
                    return [False, btstr, etstr, imname + '.fits']
            except Exception as e:
                if hasattr(e, 'message'):
                    print(e.message)
                else:
                    print(e)
                print('error in registering image: ' + btstr)
                return [False, btstr, etstr, imname + '.image']
    else:
        if os.path.exists(imname + '.image'):
            return [True, btstr, etstr, imname + '.image']
        else:
            return [False, btstr, etstr, imname + '.image']


def ptclean(vis, imageprefix, imagesuffix, ncpu, twidth, doreg, usephacenter, reftime, toTb, sclfactor, subregion,
            docompress,
            overwrite, selectdata, field, spw, timerange, uvrange, antenna, scan, observation, intent, datacolumn,
            imsize, cell, phasecenter,
            stokes, projection, startmodel, specmode, reffreq, nchan, start, width, outframe, veltype, restfreq,
            interpolation, gridder, facets, chanchunks, wprojplanes, vptable, usepointing, mosweight, aterm, psterm,
            wbawp, conjbeams, cfcache, computepastep, rotatepastep, pblimit, normtype, deconvolver, scales, nterms,
            smallscalebias, restoration, restoringbeam, pbcor, outlierfile, weighting, robust, npixels, uvtaper, niter,
            gain, threshold, nsigma, cycleniter, cyclefactor, minpsffraction, maxpsffraction, interactive, usemask,
            mask, pbmask, sidelobethreshold, noisethreshold, lownoisethreshold, negativethreshold, smoothfactor,
            minbeamfrac, cutthreshold, growiterations, dogrowprune, minpercentchange, verbose, restart, savemodel,
            calcres, calcpsf, parallel):
    if not (type(ncpu) is int):
        casalog.post('ncpu should be an integer')
        ncpu = 8

    if doreg:
        # check if ephem and msinfo exist. If not, generate one on the fly
        try:
            ephem = hf.read_horizons(vis=vis)
        except ValueError:
            print("error in obtaining ephemeris")
        try:
            msinfo = hf.read_msinfo(vis)
        except ValueError:
            print("error in getting ms info")
    else:
        ephem = None
        msinfo = None

    if imageprefix:
        workdir = os.path.dirname(imageprefix)
    else:
        workdir = './'
    tmpdir = workdir + '/tmp/'
    if not os.path.exists(tmpdir):
        os.makedirs(tmpdir)
    # get number of time pixels
    ms.open(vis)
    ms.selectinit()
    timfreq = ms.getdata(['time', 'axis_info'], ifraxis=True)
    tim = timfreq['time']
    ms.close()

    if twidth < 1:
        casalog.post('twidth less than 1. Change to 1')
        twidth = 1

    if twidth > len(tim):
        casalog.post(
            'twidth greater than # of time pixels in the dataset. Change to the timerange of the entire dateset')
        twidth = len(tim)
    # find out the start and end time index according to the parameter timerange
    # if not defined (empty string), use start and end from the entire time of the ms
    if not timerange:
        btidx = 0
        etidx = len(tim) - 1
    else:
        try:
            (tstart, tend) = timerange.split('~')
            bt_s = qa.convert(qa.quantity(tstart, 's'), 's')['value']
            et_s = qa.convert(qa.quantity(tend, 's'), 's')['value']
            # only time is given but not date, add the date (at 0 UT) from the first record
            if bt_s < 86400. or et_s < 86400.:
                bt_s += np.fix(qa.convert(qa.quantity(tim[0], 's'), 'd')['value']) * 86400.
                et_s += np.fix(qa.convert(qa.quantity(tim[0], 's'), 'd')['value']) * 86400.
            btidx = np.argmin(np.abs(tim - bt_s))
            etidx = np.argmin(np.abs(tim - et_s))
            # make the indice back to those bracket by the timerange
            if tim[btidx] < bt_s:
                btidx += 1
            if tim[etidx] > et_s:
                etidx -= 1
            if etidx <= btidx:
                print("ending time must be greater than starting time")
                print("reinitiating to the entire time range")
                btidx = 0
                etidx = len(tim) - 1
        except ValueError:
            print("keyword 'timerange' has a wrong format")

    btstr = qa.time(qa.quantity(tim[btidx], 's'), prec=9, form='fits')[0]
    etstr = qa.time(qa.quantity(tim[etidx], 's'), prec=9, form='fits')[0]

    iterable = range(btidx, etidx + 1, twidth)
    print('First time pixel: ' + btstr)
    print('Last time pixel: ' + etstr)
    print(str(len(iterable)) + ' images to clean...')

    res = []
    # partition
    clnpart = partial(clean_iter, tim, vis, imageprefix, imagesuffix,
                      twidth, doreg, docompress, usephacenter, reftime, ephem, msinfo, toTb, sclfactor, overwrite,
                      selectdata,
                      field, spw,
                      uvrange, antenna, scan, observation, intent, datacolumn, imsize, cell, phasecenter, stokes,
                      projection, startmodel, specmode, reffreq, nchan, start, width, outframe, veltype, restfreq,
                      interpolation, gridder, facets, chanchunks, wprojplanes, vptable, usepointing, mosweight, aterm,
                      psterm, wbawp, conjbeams, cfcache, computepastep, rotatepastep, pblimit, normtype, deconvolver,
                      scales, nterms, smallscalebias, restoration, restoringbeam, pbcor, outlierfile, weighting, robust,
                      npixels, uvtaper, niter, gain, threshold, nsigma, cycleniter, cyclefactor, minpsffraction,
                      maxpsffraction, interactive, usemask, mask, pbmask, sidelobethreshold, noisethreshold,
                      lownoisethreshold, negativethreshold, smoothfactor, minbeamfrac, cutthreshold, growiterations,
                      dogrowprune, minpercentchange, verbose, restart, savemodel, calcres, calcpsf, parallel, subregion,
                      tmpdir)
    timelapse = 0
    t0 = time()
    # parallelization
    if ncpu > 1:
        import multiprocessing as mprocs
        casalog.post('Perform clean in parallel ...')
        print('Perform clean in parallel ...')
        pool = mprocs.Pool(ncpu)
        res = pool.map(clnpart, iterable)
        pool.close()
        pool.join()
    else:
        casalog.post('Perform clean in single process ...')
        print('Perform clean in single process ...')
        for i in iterable:
            res.append(clnpart(i))

    t1 = time()
    timelapse = t1 - t0
    print('It took %f secs to complete' % timelapse)
    # repackage this into a single dictionary
    results = {'Succeeded': [], 'BeginTime': [], 'EndTime': [], 'ImageName': []}
    for r in res:
        results['Succeeded'].append(r[0])
        results['BeginTime'].append(r[1])
        results['EndTime'].append(r[2])
        results['ImageName'].append(r[3])

    if os.path.exists(tmpdir):
        os.system('rm -rf ' + tmpdir)

    return results



# __init__.py of suncasa.suncasatasks
import warnings

# Always available modules
from .ptclean6 import ptclean6
from .subvs import subvs
from .concateovsa import concateovsa

__all__ = ['ptclean6', 'subvs', 'concateovsa']

# Conditional import for modules that depend on 'aipy-eovsa'
try:
    from .importeovsa import importeovsa
    __all__.append('importeovsa')
except ImportError:
    warnings.warn(
        "Failed to import 'importeovsa'. This module requires 'eovsapy'. "
        "Please install 'eovsapy' to use 'importeovsa'."
    )

try:
    from .calibeovsa import calibeovsa
    __all__.append('calibeovsa')
except ImportError:
    warnings.warn(
        "Failed to import 'calibeovsa'. This module requires 'eovsapy'. "
        "Please install 'eovsapy' to use 'calibeovsa'."
    )



from suncasa.suncasatasks import importeovsa
from suncasa.suncasatasks import calibeovsa
import sys
# import importeovsa
# import calibeovsa
from casatasks import split
from eovsapy.read_idb import get_trange_files
from eovsapy.util import Time
from eovsapy import util
import numpy as np
import os
from eovsapy import dump_tsys as dt



def import_calib_idb(trange, workdir=None, ncpu=1, timebin='0s', width=1):

    """
    Script to import and calibrate IDB data based on an input time range
    Parameters
    ----------
    trange: [begin_time, end_time], in eovsa.util Time format.
        Example: Time(['2022-11-12 17:55:00', '2022-11-12 18:10:00'])
    workdir: specify where the working directory is. Default to current path
    Returns
    -------
    vis_out: concatenated CASA measurement set with initial gain, amplitude, and phase calibrations applied
    """
    if not workdir:
        workdir = os.getcwd()
        if workdir[-1] != '/':
            workdir += '/'
    if os.path.exists(workdir) == False:
        os.makedirs(workdir)


    os.chdir(workdir)

    namesuffix = '_' + trange[0].to_datetime().strftime('%H%M') + '-' + trange[1].to_datetime().strftime('%H%M')

    idbdir = util.get_idbdir(trange[0])

    info = dt.rd_fdb(Time(trange[0].to_datetime().strftime('%Y-%m-%d')))
    idxs, = np.where(info['SOURCEID'] == '')
    for k, v in info.items():
        info[k] = info[k][~(info[k] == '')]
    # sidx = np.where(
    #     np.logical_and(info['SOURCEID'] == 'Sun', info['PROJECTID'] == 'NormalObserving') & np.logical_and(
    #         info['ST_TS'].astype(np.float_) >= trange[0].lv,
    #         info['ST_TS'].astype(np.float_) <= trange[1].lv))
    
    term1 = np.logical_and(info['SOURCEID'] == 'Sun', info['PROJECTID'] == 'NormalObserving')
    st_ts_float = info['ST_TS'].astype(np.float_)

    term2_ind = []
    term3_ind = []
    for i in range(len(st_ts_float) - 1):
        if st_ts_float[i] <= trange[0].lv <= st_ts_float[i + 1]:
            term2_ind.append(i)
        if st_ts_float[i] <= trange[1].lv <= st_ts_float[i + 1]:
            term3_ind.append(i)
    if term2_ind == [] or term3_ind == []:
        print("Error: No available IDB list during the indicated time range.")

    sidx = list(range(term2_ind[0], term3_ind[0] + 1))
    # Filter out indices where term1 is not true
    sidx = [i for i in sidx if term1[i]]
    if sidx == []:
        print("Error: No available IDB list for the Sun during the indicated time range.")


    filelist = info['FILE'][sidx]
    # filelist = filelist[np.array([0, 2, 3, 4, 5, 6, 7, 8])]
    # filelist = filelist[np.array([0, 2])]

    outpath = 'msdata/'
    if not os.path.exists(outpath):
        os.makedirs(outpath)
    inpath = idbdir + '{}/'.format(trange[0].datetime.strftime("%Y%m%d"))
    ncpu = 1

    msfiles = importeovsa(idbfiles=[inpath + ll for ll in filelist], ncpu=ncpu, timebin=timebin, width=width,
                                       visprefix=outpath,
                                       nocreatms=False, doconcat=False,
                                       modelms="", doscaling=False, keep_nsclms=False, udb_corr=True,
                                       use_exist_udbcorr=True)
    # msfiles = [outpath + ll + '.ms' for ll in filelist]
    concatvis = os.path.basename(msfiles[0])[:11] + namesuffix + '.ms'
    vis = calibeovsa(msfiles, caltype=['refpha', 'phacal'], interp='nearest', doflag=True, flagant='13~15',
                                doimage=False, doconcat=True,
                                concatvis=concatvis, keep_orig_ms=False)
    
                   

    return concatvis


